diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/Bug.pm bugzilla-2.17.4-new/Bug.pm
--- bugzilla-2.17.4/Bug.pm	Wed Apr  9 02:29:15 2003
+++ bugzilla-2.17.4-new/Bug.pm	Tue Jul  8 12:47:57 2003
@@ -40,6 +40,7 @@
 use Bugzilla::FlagType;
 use Bugzilla::User;
 use Bugzilla::Util;
+use Bugzilla::CustomFields qw(get_custom_fields);
 
 sub fields {
     # Keep this ordering in sync with bugzilla.dtd
@@ -51,6 +52,7 @@
                     priority bug_severity target_milestone
                     dependson blocked votes
                     reporter assigned_to cc
+                    custom_fields
                    );
 
     if (Param('useqacontact')) {
@@ -258,6 +260,10 @@
   if (@blocked) {
     $self->{'blocked'} = \@blocked;
   }
+
+  $self->{'custom_fields'} = [
+    get_custom_fields($self->{'product_id'}, $self->{'bug_id'})
+  ];
 
   return $self;
 }
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/Bugzilla/CustomFields/Util.pm bugzilla-2.17.4-new/Bugzilla/CustomFields/Util.pm
--- bugzilla-2.17.4/Bugzilla/CustomFields/Util.pm	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/Bugzilla/CustomFields/Util.pm	Tue Jul  8 16:44:11 2003
@@ -0,0 +1,200 @@
+package Bugzilla::CustomFields::Util;
+
+use Date::Parse;
+use Exporter;
+use strict;
+use vars qw(@ISA @EXPORT);
+
+@ISA    = qw(Exporter);
+@EXPORT = qw(sql_quote send_sql process_query simple_query last_insert_id
+             parse_integer parse_date stringify_date is_numeric);
+
+my %escape = (
+	"\n" => '\\n',
+	"\r" => '\\r',
+	"'"  => '\\\'',
+	"\\" => '\\\\',
+	"\0" => '\\0',
+);
+
+my $echars = join "", keys %escape;
+
+#
+#  sql_quote($str)
+#
+#  Returns the quoted SQL representation of the string $string.  Newline,
+#  carriage return, single quote, backslash, and null characters are
+#  replaced with escape sequences; the string is untainted; and finally,
+#  the string is surrounded with single quotes and returned.
+#
+
+sub sql_quote {
+	my $str = shift;
+	$str =~ s/([\Q$echars\E])/$escape{$1}/go;
+    ::trick_taint($str);
+    return "'$str'";
+}
+
+#
+#  send_sql($sql)
+#
+#  This routine massages the SQL statement contained in $sql to make it
+#  more suitable for logging purposes, then passes it on to SendSQL().
+#  Leading and trailing whitespace is removed, and all newline-indentation
+#  sequences are collapsed to a single space.
+#
+
+sub send_sql {
+    my $sql = shift;
+    $sql = ::trim($sql);
+    $sql =~ s/\n\s*/ /g;
+    ::SendSQL($sql);
+}
+
+#
+#  process_query($sql [, $sub])
+#
+#  This routine iterates over the result set obtained by passing $sql to
+#  send_sql(), passing each row in the set to the subroutine $sub.  The
+#  number of rows in the result set is returned.
+#
+
+sub process_query {
+    my ($sql, $sub) = @_;
+    my $count = 0;
+    my @row;
+
+    send_sql($sql);
+
+    while (::MoreSQLData()) {
+        @row = ::FetchSQLData();
+        $sub->(@row) if $sub;
+        ++$count;
+    }
+
+    return $count;
+
+}
+
+#
+#  simple_query($sql)
+#
+#  This routine is useful for queries which return only a single column.
+#  It returns a list containing the first column of each row in the result
+#  set of the SQL statement $sql.
+#
+
+sub simple_query {
+    my $sql = shift;
+    my @result;
+
+    send_sql($sql);
+
+    while (::MoreSQLData()) {
+        push @result, ::FetchOneColumn();
+    }
+
+    return wantarray ? @result : $result[0];
+
+}
+
+#
+#  last_insert_id()
+#
+#  Returns the integer ID created by the last INSERT into a table with an
+#  auto_increment column.
+#
+
+sub last_insert_id {
+    send_sql('SELECT LAST_INSERT_ID()');
+    return ::FetchOneColumn();
+}
+
+#
+#  parse_integer($str)
+#
+#  Attempts to parse a string containing an integer.  Leading and trailing
+#  whitespace is trimmed; what is left must resemble the usual integer
+#  pattern, otherwise an exception is thrown.  The trimmed string is
+#  returned.
+#
+
+sub parse_integer {
+    my $str = shift;
+
+    $str = ::trim($str);
+
+    if ($str eq "") {
+        $str = undef;
+    } elsif ($str =~ /^([-+]?\d+)$/) {
+		$str = $1;
+	} else {
+        die qq(Invalid integer value "$str"\n);
+    }
+
+    return $str;
+
+}
+
+#
+#  stringify_date($str, $withtime)
+#
+#  Attempts to parse a string containing a date.  Leading and trailing
+#  whitespace is trimmed; what is left must be parseable by Date::Parse, or
+#  an exception is thrown.  Otherwise, the parsed date is cast in the format
+#  YYYYMMDDHHMMSS, which is returned.  If the original string contained no
+#  non-whitespace characters, undef is returned instead.
+#
+
+sub stringify_date {
+    my ($str, $withtime) = @_;
+	my $result;
+
+	my $date = parse_date($str, $withtime);
+
+	if (@$date) {
+        $result = sprintf '%04d%02d%02d%02d%02d%02d', @$date;
+	}
+
+    return $result;
+
+}
+
+#
+#  parse_date($str, $withtime)
+#
+#  Returns an internal representation of the date in $str.  The time is
+#  included if $withtime is true.  The internal representation is undef if
+#  $str has no non-whitespace characters; otherwise it is a reference to an
+#  array of the form [ $year, $month, $day (, $hour, $minute, $second) ].
+#
+
+sub parse_date {
+	my ($str, $withtime) = @_;
+	my $date;
+
+	$str = ::trim($str);
+	
+	if ($str ne "") {
+        my ($sec, $min, $hour, $day, $month, $year) = strptime($str)
+            or die qq(Invalid date value "$str"\n);
+		$date = [ $year + 1900, $month + 1, $day ];
+		push @$date, $hour, $min, $sec if $withtime;
+	}
+
+	return $date;
+
+}
+
+#
+#  is_numeric($str)
+#
+#  Returns true if $str is of nonzero length and consists solely of numeric
+#  characters, and false otherwise.
+#
+
+sub is_numeric {
+    my $str = shift;
+    $str =~ /^\d+$/;
+}
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/Bugzilla/CustomFields.pm bugzilla-2.17.4-new/Bugzilla/CustomFields.pm
--- bugzilla-2.17.4/Bugzilla/CustomFields.pm	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/Bugzilla/CustomFields.pm	Tue Jul  8 16:50:56 2003
@@ -0,0 +1,913 @@
+package Bugzilla::CustomFields;
+
+use Bugzilla::CustomFields::Util;
+use Bugzilla::Util;
+use Exporter;
+use strict;
+use vars qw(@ISA @EXPORT @EXPORT_OK);
+
+@ISA       = qw(Exporter);
+@EXPORT_OK = qw(get_custom_fields check_custom_fields store_custom_fields
+                update_custom_fields form_to_sql all_fields_by_product
+                read_cf_data class_for_type table_for_type
+                read_default_field_data read_selection_metadata
+                classify_fields insert_field_data);
+
+#  %class is a mapping from single-character database field type to a
+#  textual "class".  Field types in the same class keep custom field data
+#  in the same database table.
+
+my %class = (
+    i => 'integer',
+    d => 'date',
+    t => 'date',
+    s => 's_string',
+    l => 'l_string',
+    e => 'selection',
+    m => 'selection',
+);
+
+#  %classinfo maps the classes defined above to several bits of information.
+#
+#  'table':     The database table that stores bug data for the class
+#  'column':    The column within the table where the bug data resides
+#  'quote':     Boolean; true if field data should be quoted (via
+#               SqlQuote()) when inserted into the table, false otherwise
+#  'multiple':  Boolean; true if the field data is a list, false if the
+#               data is a scalar
+#  'parse':     A reference to a routine that converts the "raw" data
+#               returned from the database into an internal representation
+#  'compare':   A reference to a routine that compares the internal
+#               representations of two data
+#  'formsub':   A reference to a routine that accepts a custom field hash
+#               as an argument and returns an internal representation of
+#               that field obtained by examining the appropriate CGI
+#               parameters.
+#  'as_string': A reference to a routine that accepts an internal
+#               representation of a field and returns a simple string
+#               representation of that field
+#  'as_sql':    A reference to a routine that accepts an internal
+#               representation of a field and returns the SQL
+#               representation of that field
+#  'update':    A reference to a routine that updates the database for
+#               fields of that class
+#
+
+my %classinfo = (
+    integer => {
+        table     => 'cf_integer',
+        column    => 'value',
+        quote     => 0,
+        multiple  => 0,
+		parse     => \&process_null,
+		compare   => \&compare_integer,
+		formsub   => \&read_form_integer,
+		as_string => \&simple_scalar,
+		as_sql    => \&integer_as_sql,
+		update    => \&update_nonselection_fields,
+    },
+    date => {
+        table     => 'cf_date',
+        column    => 'value',
+        quote     => 1,
+        multiple  => 0,
+		parse     => \&process_date,
+		compare   => \&compare_date,
+		formsub   => \&read_form_date,
+		as_string => \&date_as_string,
+		as_sql    => \&date_as_sql,
+		update    => \&update_nonselection_fields,
+    },
+    s_string => {
+        table     => 'cf_shortstring',
+        column    => 'value',
+        quote     => 1,
+        multiple  => 0,
+		parse     => \&process_null,
+		compare   => \&compare_string,
+		formsub   => \&read_form_string,
+		as_string => \&simple_scalar,
+		as_sql    => \&string_as_sql,
+		update    => \&update_nonselection_fields,
+    },
+    l_string => {
+        table     => 'cf_longstring',
+        column    => 'value',
+        quote     => 1,
+        multiple  => 0,
+		compare   => \&compare_string,
+		parse     => \&process_null,
+		formsub   => \&read_form_string,
+		as_string => \&simple_scalar,
+		as_sql    => \&string_as_sql,
+		update    => \&update_nonselection_fields,
+    },
+    selection => {
+        table     => 'cf_selection',
+        column    => 'label_id',
+        quote     => 0,
+        multiple  => 1,
+		formsub   => \&read_form_selection,
+		as_string => undef,
+		update    => \&update_selection_fields,
+    },
+);
+
+#
+#  get_custom-fields($product_id, $bug_id)
+#
+#  Returns a list of all custom fields which belong to product $product_id,
+#  sorted into the proper order.  If the $bug_id argument is present and
+#  defined, each field's "value" member is set.
+#
+
+sub get_custom_fields {
+	my ($product_id, $bug_id) = @_;
+
+	is_numeric($product_id) or die qq(Invalid product ID "$product_id"\n);
+	!defined $bug_id or is_numeric($bug_id)
+		or die qq(Invalid bug ID "$bug_id"\n);
+
+	my @field_id = simple_query(qq{
+		SELECT gm.field_id
+		FROM cf_product_members pm
+		  INNER JOIN cf_group_members gm USING (group_id)
+		WHERE pm.product_id = $product_id
+		ORDER BY pm.sortkey, gm.sortkey
+	});
+
+	my @fields = read_custom_fields(@field_id);
+
+	if (defined $bug_id) {
+		read_bug_data($bug_id, @fields);
+		set_label_text(\@fields, 'value');
+	}
+
+	return @fields;
+
+}
+
+
+#
+#  read_custom_fields(@field_id)
+#
+#  Creates a new field hash for each of the fields enumerated in @field_id.
+#  Returns a list of the new hashes.
+#
+
+sub read_custom_fields {
+    my @field_id = @_;
+    my $field_set = join(',', @field_id);
+
+    send_sql(qq{
+        SELECT field_id, field_name, display_name, field_type,
+               required, selection_id
+        FROM custom_fields
+        WHERE field_id IN ($field_set)
+    });
+
+	my %field;
+
+    while (::MoreSQLData()) {
+		my $field = read_cf_data();
+		$field{ $field->{id} } = $field;
+    }
+
+	my @fields = map $field{$_}, @field_id;
+
+    read_default_field_data(@fields);
+    read_selection_metadata(@fields);
+
+    return wantarray ? @fields : $fields[0];
+
+}
+
+#
+#  read_cf_data()
+#
+#  Creates a field hash and returns it.  A SELECT statement must previously
+#  have been sent to the database which returns field information; the
+#  first through sixth columns must be an ID, a name, a display name, a
+#  type, a "required" flag, and a selection ID, in that order.  Any extra
+#  columns are returned after the new field hash in the return list.  If the
+#  first column is NULL, undef is returned.
+#
+
+sub read_cf_data {
+    my ($id, $name, $dname, $type, $required, $selection_id, @extra) =
+        ::FetchSQLData();
+    my $field;
+
+    if (defined $id) {
+        defined(my $class = $class{$type})
+            or die qq(Database corrupted: Field "$name" ) .
+			       qq(has unknown type "$type"\n);
+        $field = {
+            id           => $id,
+            name         => $name,
+            display_name => $dname,
+            type         => $type,
+            required     => $required,
+            extended     => $type eq 'l' || $type eq 'm' || $type eq 't',
+            selection_id => $selection_id,
+            is_integer   => $class eq 'integer',
+            is_date      => $class eq 'date',
+            is_string    => $class eq 'l_string' || $class eq 's_string',
+            is_selection => $class eq 'selection',
+			as_string    => $classinfo{$class}{as_string},
+        };
+    }
+
+    return wantarray ? ($field, @extra) : $field;
+
+}
+
+#
+#  read_bug_data($bug_id, @fields)
+#
+#  This routine reads the value of each field in @fields for bug $bug_id.
+#  If $bug_id is defined, the value of each field for that bug is stored
+#  in the 'value' entry of each field's hash; otherwise, the default value
+#  of the field is stored in the 'default' entry of each hash.
+#
+
+sub read_bug_data {
+    my ($bug_id, @fields) = @_;
+	my $name = defined $bug_id ? 'value' : 'default';
+    my %by_class = classify_fields(@fields);
+
+    while (my ($class, $fields) = each %by_class) {
+        my @id = map $$_{id}, @$fields;
+        my @values = read_field_data($bug_id, $class, \@id);
+
+        for (my $i = 0; $i < @values; ++$i) {
+            $fields->[$i]{$name} = $values[$i];
+        }
+
+        #  Special case for dates:
+        if ($class eq 'date') {
+            for my $field (grep !$$_{extended}, @$fields) {
+				splice @{ $field->{$name} }, 3, 3;
+            }
+        }
+
+    }
+}
+
+#
+#  read_default_field_data(@fields)
+#
+#  This routine stores the default value of each field in @fields.
+#
+
+sub read_default_field_data {
+	my @fields = @_;
+	read_bug_data(undef, @fields);
+}
+
+#
+#  read_selection_metadata(@fields)
+#
+#  This routine stores information specific to selection fields in each
+#  of the selection fields in @fields.  The extra elements are 'unset'
+#  (a field's "unset" label), 'label' (a mapping from label_id to the text
+#  and order key of that label), and 'domain' (an array of all labels in
+#  the field's domain, sorted into the proper order).  In addition, the
+#  'default' element of each field hash is reset to refer to each labels'
+#  text, rather than their numeric IDs.
+#
+
+sub read_selection_metadata {
+    my @fields = @_;
+    my %by_class = classify_fields(@fields);
+    defined(my $fields = $by_class{selection}) or return;
+
+    my (%seen, %unset, %label, %lblid);
+
+    my @sel_ids = grep !$seen{$_}++, map $$_{selection_id}, @$fields;
+    my $id_set  = join(',', @sel_ids);
+
+    process_query(qq{
+        SELECT selection_id, unset_label
+        FROM cf_selections
+        WHERE selection_id IN ($id_set)
+    }, sub {
+        $unset{ $_[0] } = $_[1];
+    });
+
+    process_query(qq{
+        SELECT selection_id, label_id, label, sortkey
+        FROM cf_selection_labels
+        WHERE selection_id IN ($id_set)
+    }, sub {
+        $label{ $_[0] }{ $_[1] } = { text => $_[2], order => $_[3] };
+		$lblid{ $_[0] }{ $_[2] } = $_[1];
+    });
+
+    for my $field (@$fields) {
+        my $selection_id = $field->{selection_id};
+        my $label = $field->{label} = $label{$selection_id};
+		$field->{lblid}  = $lblid{$selection_id};
+        $field->{unset}  = $unset{$selection_id};
+		$field->{domain} = [ keys %$label ];
+    }
+
+	set_label_text($fields, 'domain', 'default');
+
+}
+
+#
+#  set_label_text($fields, @keys)
+#
+#  This routine converts selection field values from unsorted lists of
+#  integer selection label IDs to sorted lists of textual labels.  Each of
+#  the hash entries in @keys in each field in @$fields is so modified.
+#
+
+sub set_label_text {
+	my ($fields, @keys) = @_;
+
+	for my $field (@$fields) {
+		next unless $field->{is_selection};
+		my $label = $field->{label};
+		for my $key (@keys) {
+			my $array = $field->{$key};
+			@$array = map  { $_->{text} }
+			          sort { $a->{order} <=> $b->{order} }
+			          @$label{ @$array };
+		}
+	}
+
+}
+
+#
+#  classify_fields(@fields)
+#
+#  Iteratates over the fields in @fields, storing each one in a hash keyed
+#  by the class of each.  Returns the flattened hash.
+#
+
+sub classify_fields {
+    my @fields = @_;
+    my %by_class;
+
+    for my $field (@fields) {
+        my $class = $class{ $field->{type} };
+        push @{ $by_class{$class} }, $field;
+    }
+
+    return %by_class;
+}
+
+#
+#  read_field_data($bug_id, $class, $field_ids)
+#
+#  This routine reads field data for the bug $bug_id for all fields of a
+#  given class, namely $class.  $field_ids is a reference to an array of
+#  numeric field IDs.
+#
+
+sub read_field_data {
+    my ($bug_id, $class, $field_ids) = @_;
+    my $info = $classinfo{$class};
+    my ($table, $column, $multiple, $parse) =
+		@$info{qw(table column multiple parse)};
+    my $field_set = join(',', @$field_ids);
+	my $bug_clause = defined $bug_id ? "bug_id = $bug_id" : "bug_id IS NULL";
+    my %value;
+
+    @value{ @$field_ids } = map [ ], 0 .. $#$field_ids;
+
+	my $sql = qq{
+		SELECT field_id, $column
+		FROM $table
+		WHERE $bug_clause AND field_id IN ($field_set)
+	};
+
+	if ($multiple) {
+		process_query($sql, sub {
+			push @{ $value{ $_[0] } }, $_[1];
+		});
+	} else {
+		process_query($sql, sub {
+			$value{ $_[0] } = $parse->($_[1]);
+		});
+	}
+
+    return @value{ @$field_ids };
+
+}
+
+#
+#  check_custom_fields(@fields)
+#
+#  This routine checks that valid CGI parameters exist for each of the
+#  fields in @fields.  It throws a CodeError if any field has not been
+#  validly set.
+#
+
+sub check_custom_fields {
+	my @fields = @_;
+
+	for my $field (@fields) {
+		my $name = $field->{name};
+		my $valid = 1;
+		if (exists $::FORM{$name}) {
+			if ($field->{is_string}) {
+				my $maxlen = $field->{extended} ? 65535 : 255;
+				$valid = length($::FORM{$name}) <= $maxlen;
+			} elsif ($field->{is_integer}) {
+				eval { parse_integer($::FORM{$name}) };
+				$valid = not $@;
+			} elsif ($field->{is_date}) {
+				eval { parse_date($::FORM{$name}, $field->{extended}) };
+				$valid = not $@;
+			} elsif ($field->{is_selection}) {
+				my $form = $::MFORM{$name};
+				if (!$field->{extended} && @$form != 1) {
+					$valid = 0;
+				} else {
+					$valid = 1;
+					unless (!$field->{extended}
+							&& $form->[0] eq $field->{unset}) {
+						for my $selection (@$form) {
+							grep($selection eq $_, @{ $field->{domain} })
+								or $valid = 0;
+						}
+					}
+				}
+			}
+		}
+
+		unless ($valid) {
+			$::vars->{'field'} = $name;
+			::ThrowCodeError('illegal_field', undef, 'abort');
+		}
+
+	}
+
+}
+
+#
+#  insert_field_data($class, $data, @bugs)
+#
+#  This routine inserts field data into a single database table.  $data
+#  is a reference to a hash which associates field IDs (the keys) with field
+#  data (the values).  All fields must of the same class, namely $class.  
+#  All of the indicated fields in each of the bugs enumerated in @bugs will
+#  have their fields set to the indicated values.  The string 'NULL' may
+#  occur in @bugs, in which case the default values of the fields will be
+#  set.
+#
+
+sub insert_field_data {
+    my ($class, $fields, $values, @bugs) = @_;
+    my $info = $classinfo{$class};
+    my ($table, $column, $quote, $multiple) =
+        @$info{qw(table column quote multiple)};
+	my @insert;
+
+	for (my $i = 0; $i < @$fields; ++$i) {
+		my $field = $fields->[$i];
+		my $value = $values->[$i];
+        my @value = $multiple ? @$value : $value;
+        for my $val (@value) {
+            if (defined $val) {
+                $val = sql_quote($val) if $quote;
+            } else {
+                $val = 'NULL';
+            }
+        }
+        for my $bug_id (@bugs) {
+			push @insert, map "($bug_id, $field, $_)", @value;
+        }
+    }
+
+	if (@insert) {
+		my $list = join(', ', @insert);
+		send_sql(qq{
+			INSERT INTO $table (bug_id, field_id, $column)
+			            VALUES $list
+		});
+	}
+
+}
+
+#
+#  store_custom_fields($bug_id, @fields)
+#
+#  This routine stores custom field data for each field in @fields for
+#  bug $bug_id.  The field data is taken from the CGI parameters for each
+#  field.  No error checking is performed; check_custom_fields() should be
+#  called before calling this routine.
+#
+
+sub store_custom_fields {
+	my ($bug_id, @fields) = @_;
+	my %by_class = classify_fields(@fields);
+
+	while (my ($class, $fields) = each %by_class) {
+		my $formsub = $classinfo{$class}{formsub};
+		my @fldid   = map $$_{id},        @$fields;
+		my @values  = map $formsub->($_), @$fields;
+		insert_field_data($class, \@fldid, \@values, $bug_id);
+	}
+
+}
+
+#
+#  update_custom_fields($bug_id, @fields)
+#
+#  This routine updates custom field data for each field in @fields for bug
+#  $bug_id.  The field data is taken from the CGI parameters for each
+#  field.
+#
+
+sub update_custom_fields {
+	my ($bug_id, @fields) = @_;
+	my %by_class = classify_fields(@fields);
+
+	while (my ($class, $fields) = each %by_class) {
+		$classinfo{$class}{update}->($class, $bug_id, @$fields);
+	}
+
+}
+
+#
+#  update_nonselection_fields($class, $bug_id, @fields)
+#
+#  This routine updates scalar (ie, nonselection) fields.  Each of the
+#  fields in @fields, which must all be of class $class, is updated for
+#  bug $bug_id.
+#
+
+sub update_nonselection_fields {
+	my ($class, $bug_id, @fields) = @_;
+	my $info = $classinfo{$class};
+	my ($table, $column, $compare, $as_sql, $formsub) =
+		@$info{qw(table column compare as_sql formsub)};
+	my (@id, @when);
+
+	for my $field (@fields) {
+		my $id  = $field->{id};
+		my $old = $field->{value};
+		my $new = $formsub->($field);
+		if (
+			(defined $old xor defined $new)
+				or
+			(defined $old and defined $new and $compare->($old, $new) != 0)
+		) {
+			push @id, $id;
+			push @when, "WHEN $id THEN " . $as_sql->($new);
+		}
+	}
+
+	if (@id) {
+		my $id_set    = join(', ', @id);
+		my $case_body = join(' ', @when);
+		send_sql(qq{
+			UPDATE $table
+			SET $column = CASE field_id $case_body END
+			WHERE bug_id = $bug_id AND field_id IN ($id_set)
+		});
+	}
+
+}
+
+#
+#  update_selection_fields($class, $bug_id, @fields)
+#
+#  This routine updates selection fields.  Each of the fields in @fields is
+#  updated for bug $bug_id.  The $class parameter is ignored (it will
+#  always be 'selection').
+#
+
+sub update_selection_fields {
+	my ($class, $bug_id, @fields) = @_;
+	my $info = $classinfo{$class};
+	my ($table, $column, $formsub) = @$info{qw(table column formsub)};
+	my (@insert, @delete);
+
+	for my $field (@fields) {
+		my ($field_id, $lblid, $value) = @$field{qw(id lblid value)};
+		my $old = [ @$lblid{ @$value } ];
+		my $new = $formsub->($field);
+		my (%old, %new);
+		@old{ @$old } = ();
+		@new{ @$new } = ();
+		my @added   = grep !exists $old{$_}, @$new;
+		my @removed = grep !exists $new{$_}, @$old;
+		push @insert, map "($bug_id, $field_id, $_)", @added;
+		push @delete, "(field_id = $field_id AND" .
+		              " label_id IN (" . join(', ', @removed) . "))"
+		              if @removed;
+	}
+
+	if (@insert) {
+		my $values = join(', ', @insert);
+		send_sql(qq{
+			INSERT INTO $table (bug_id, field_id, $column)
+			            VALUES $values
+		});
+	}
+
+	if (@delete) {
+		my $clause = join(' OR ', @delete);
+		send_sql(qq{
+			DELETE FROM $table
+			WHERE bug_id = $bug_id AND ($clause)
+		});
+	}
+
+}
+
+#
+#  compare_integer($a, $b)
+#
+#  Compares two integers; returns $a <=> $b.
+#
+
+sub compare_integer {
+	return $_[0] <=> $_[1];
+}
+
+#
+#  compare_string($a, $b)
+#
+#  Compares two strings; returns $a cmp $b.
+#
+
+sub compare_string {
+	return $_[0] cmp $_[1];
+}
+
+#
+#  compare_date($a, $b)
+#
+#  Compares two dates $a and $b, each represented as a reference to an
+#  array of the form [ $year, $month, $day (, $hour, $minute, $second ) ].
+#  Returns -1 if $a is earlier than $b, +1 if $a is later than $b, or 0 if
+#  $a is the same as $b.
+#
+
+sub compare_date {
+	my ($a, $b) = @_;
+	my $cmp;
+
+	# @$a == @$b or die "Can't compare different date types\n";
+
+	for (my $i = 0; $i < @$a; ++$i) {
+		return $cmp if ($cmp = $a->[$i] <=> $b->[$i]) != 0;
+	}
+
+	return 0;
+
+}
+
+#
+#  read_form_integer($field)
+#
+#  Reads and returns the value of the integer field $field, as taken from
+#  the appropriate CGI parameter.
+#
+
+sub read_form_integer {
+	my $field = shift;
+	my $name  = $field->{name};
+	return parse_integer($::FORM{$name});
+}
+
+#
+#  read_form_date($field)
+#
+#  Reads and returns the value of the date field $field, as taken from the
+#  appropriate CGI parameter.
+#
+
+sub read_form_date {
+	my $field = shift;
+	my ($name, $extended) = @$field{qw(name extended)};
+	return parse_date($::FORM{$name}, $extended);
+}
+
+#
+#  read_form_string($field)
+#
+#  Reads and returns the value of the string field $field, as taken from
+#  the appropriate CGI parameter.
+#
+
+sub read_form_string {
+	my $field   = shift;
+	my $name    = $field->{name};
+	my $formval = $::FORM{$name};
+	trick_taint($formval);
+	$formval =~ s/\r\n?/\n/g;
+	return $formval;
+}
+
+#
+#  read_form_selection($field)
+#
+#  Reads and returns the value of the selection field $field, as taken from
+#  the appropriate CGI parameter.
+#
+
+sub read_form_selection {
+	my $field = shift;
+	my ($name, $extended, $unset, $lblid) =
+		@$field{qw(name extended unset lblid)};
+	my $formval = $::MFORM{$name};
+	my @value;
+
+	if ($formval) {
+		unless (!$extended && @$formval == 1 && $formval->[0] eq $unset) {
+			@value = @$lblid{ @$formval };
+		}
+	}
+
+	return \@value;
+
+}
+
+#
+#  process_null($a)
+#
+#  Returns $a.
+#
+
+sub process_null {
+	return $_[0];
+}
+
+#
+#  process_date($text, $type)
+#
+#  Converts the string $text into the internal representation of a date.
+#
+
+sub process_date {
+	my ($text, $type) = @_;
+	my $result;
+
+	my @num = $text =~ /(\d+)/g;
+	if (@num) {
+		$_ += 0 for @num;
+		$result = \@num;
+	}
+
+	return $result;
+
+}
+
+#
+#  simple_scalar($a)
+#
+#  Returns $a.
+#
+
+sub simple_scalar {
+	return $_[0];
+}
+
+#
+#  date_as_string($date)
+#
+#  Returns a string representation of the date $date, which is in the form
+#  [ $year, $month, $day (, $hour, $minute, $second) ].
+#
+
+sub date_as_string {
+	my $date = shift;
+	my $str = "";
+	if ($date) {
+		my ($year, $month, $day, $hour, $minute, $second) = @$date;
+		$str .= sprintf '%04d-%02d-%02d', $year, $month, $day if @$date >= 3;
+		$str .= sprintf ' %02d:%02d:%02d', $hour, $minute, $second if @$date == 6;
+	}
+	return $str;
+}
+
+#
+#  form_to_sql($text, $type)
+#
+#  This routine returns the SQL representation of $text, which is assumed
+#  to describe a field of type $type.
+#
+
+sub form_to_sql {
+	my ($text, $type) = @_;
+	defined(my $class = $class{$type}) or die qq(Unknown type "$type"\n);
+	my $info = $classinfo{$class};
+	my ($parse, $as_sql) = @$info{qw(parse as_sql)};
+	return $as_sql->($parse->($text, $type));
+}
+
+#
+#  integer_as_sql($int)
+#
+#  Returns the SQL representation of the integer $int.
+#
+
+sub integer_as_sql {
+	my $int = shift;
+	return defined $int ? $int : 'NULL';
+}
+
+#
+#  string_as_sql($str)
+#
+#  Returns the SQL representation of the string $str.
+#
+
+sub string_as_sql {
+	my $str = shift;
+	return defined $str ? sql_quote($str) : 'NULL';
+}
+
+#
+#  date_as_sql($date)
+#
+#  Returns the SQL representation of the date $date.
+#
+
+sub date_as_sql {
+	my $date = shift;
+	return defined $date ? sql_quote(date_as_string($date)) : 'NULL';
+}
+
+#
+#  all_fields_by_product()
+#
+#  (Documentation forthcoming.)
+#
+
+sub all_fields_by_product {
+	my (@all, %field_by_id, %product_by_id, %field, %product, %seen_field);
+
+	send_sql(qq{
+		SELECT field_id, field_name, display_name, field_type,
+		       required, selection_id
+		FROM custom_fields
+	});
+
+	while (::MoreSQLData()) {
+		my $field = read_cf_data();
+		$field_by_id{ $field->{id} } = $field;
+		$field{ $field->{name} } = $field;
+	}
+
+	process_query(qq{
+		SELECT id, name
+		FROM products
+	}, sub {
+		$product_by_id{ $_[0] } = $_[1];
+		$product{ $_[1] } = [ ];
+	});
+
+	process_query(qq{
+		SELECT pm.product_id, gm.field_id
+		FROM cf_product_members pm
+		  INNER JOIN cf_group_members gm USING (group_id)
+		  INNER JOIN custom_fields cf USING (field_id)
+		ORDER BY pm.product_id, pm.sortkey, gm.sortkey
+	}, sub {
+		my $product_name = $product_by_id{ $_[0] };
+		my $field_name   =   $field_by_id{ $_[1] };
+		push @{ $product{$product_name} }, $field_name;
+		if (!$seen_field{$field_name}++) {
+			push @all, $field_name;
+		}
+	});
+
+	return (\@all, \%product, \%field);
+
+}
+
+#
+#  class_for_type($type)
+#
+#  Returns the class to which fields of type $type belong.
+#
+
+sub class_for_type {
+	my $type = shift;
+	defined(my $class = $class{$type}) or die qq(Unknown type "$type"\n);
+	return $class;
+}
+
+#
+#  table_for_type($type)
+#
+#  Returns the table which holds data for fields of type $type.
+#
+
+sub table_for_type {
+	my $type = shift;
+	defined(my $class = $class{$type}) or die qq(Unknown type "$type"\n);
+	return $classinfo{$class}{table};
+}
+
+
+1;
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/bz-cf-data.txt bugzilla-2.17.4-new/bz-cf-data.txt
--- bugzilla-2.17.4/bz-cf-data.txt	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/bz-cf-data.txt	Tue Jul  8 17:00:56 2003
@@ -0,0 +1,64 @@
+INSERT INTO custom_fields VALUES
+(1, 'short_string_field', 'Short String Field', 's', 1, NULL),
+(2, 'long_string_field', 'Long String Field', 'l', 0, NULL),
+(3, 'ssel_field', 'Single Selection Field', 'e', 0, 1),
+(4, 'msel_field', 'Multiselection Field', 'm', 0, 2),
+(5, 'int_field', 'Integer Field', 'i', 0, NULL),
+(6, 'date_field', 'Date Field', 'd', 0, NULL),
+(7, 'date_time_field', 'Date/Time Field', 't', 0, NULL);
+
+INSERT INTO cf_groups VALUES
+(1, 'String Group'),
+(2, NULL),
+(3, NULL),
+(4, NULL),
+(5, 'Date Group');
+
+INSERT INTO cf_product_members VALUES
+(1, 1, 1),
+(1, 2, 2),
+(1, 3, 3),
+(1, 4, 4),
+(1, 5, 5);
+
+INSERT INTO cf_group_members VALUES
+(1, 1, 1),
+(1, 2, 2),
+(2, 3, 1),
+(3, 4, 1),
+(4, 5, 1),
+(5, 6, 1),
+(5, 7, 2);
+
+INSERT INTO cf_selections VALUES
+(1, 'None'),
+(2, NULL);
+
+INSERT INTO cf_selection_labels VALUES
+(1, 1, 1, 'foo'),
+(2, 1, 2, 'bar'),
+(3, 1, 3, 'baz'),
+(4, 1, 4, 'bletch'),
+(5, 2, 1, 'one'),
+(6, 2, 2, 'two'),
+(7, 2, 3, 'three'),
+(8, 2, 4, 'four');
+
+INSERT INTO cf_integer VALUES
+(NULL, 5, 10);
+
+INSERT INTO cf_shortstring VALUES
+(NULL, 1, '');
+
+INSERT INTO cf_longstring VALUES
+(NULL, 2, '');
+
+INSERT INTO cf_date VALUES
+(NULL, 6, NULL),
+(NULL, 7, NULL);
+
+INSERT INTO cf_selection VALUES
+(NULL, 3, 1),
+(NULL, 4, 5),
+(NULL, 4, 7);
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/bz-cf-schema.txt bugzilla-2.17.4-new/bz-cf-schema.txt
--- bugzilla-2.17.4/bz-cf-schema.txt	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/bz-cf-schema.txt	Tue Jul  8 17:00:52 2003
@@ -0,0 +1,158 @@
+DROP TABLE IF EXISTS custom_fields;
+
+CREATE TABLE custom_fields (
+    field_id INTEGER AUTO_INCREMENT PRIMARY KEY,
+    field_name TINYTEXT NOT NULL,
+    display_name TINYTEXT NOT NULL,
+    field_type CHAR BINARY NOT NULL,
+    required BOOL NOT NULL,
+    selection_id INTEGER
+);
+
+DROP TABLE IF EXISTS cf_groups;
+
+CREATE TABLE cf_groups (
+    group_id INTEGER AUTO_INCREMENT PRIMARY KEY,
+    group_name TINYTEXT
+);
+
+DROP TABLE IF EXISTS cf_product_members;
+
+CREATE TABLE cf_product_members (
+    product_id SMALLINT,
+    group_id INTEGER NOT NULL,
+    sortkey INTEGER NOT NULL,
+    INDEX (product_id)
+);
+
+DROP TABLE IF EXISTS cf_group_members;
+
+CREATE TABLE cf_group_members (
+    group_id INTEGER NOT NULL,
+    field_id INTEGER NOT NULL,
+    sortkey INTEGER NOT NULL,
+    INDEX (group_id)
+);
+
+DROP TABLE IF EXISTS cf_selections;
+
+CREATE TABLE cf_selections (
+    selection_id INTEGER AUTO_INCREMENT PRIMARY KEY,
+    unset_label TINYTEXT NOT NULL
+);
+
+DROP TABLE IF EXISTS cf_selection_labels;
+
+CREATE TABLE cf_selection_labels (
+    label_id INTEGER AUTO_INCREMENT PRIMARY KEY,
+    selection_id INTEGER NOT NULL,
+    sortkey INTEGER NOT NULL,
+    label TINYTEXT NOT NULL,
+    INDEX (selection_id)
+);
+
+DROP TABLE IF EXISTS cf_integer;
+
+CREATE TABLE cf_integer (
+    bug_id INTEGER,
+    field_id INTEGER NOT NULL,
+    value INTEGER,
+    UNIQUE KEY (bug_id, field_id)
+);
+
+DROP TABLE IF EXISTS cf_shortstring;
+
+CREATE TABLE cf_shortstring (
+    bug_id INTEGER,
+    field_id INTEGER NOT NULL,
+    value TINYTEXT,
+    UNIQUE KEY (bug_id, field_id)
+);
+
+DROP TABLE IF EXISTS cf_longstring;
+
+CREATE TABLE cf_longstring (
+    bug_id INTEGER,
+    field_id INTEGER NOT NULL,
+    value TEXT,
+    UNIQUE KEY (bug_id, field_id)
+);
+
+DROP TABLE IF EXISTS cf_date;
+
+CREATE TABLE cf_date (
+    bug_id INTEGER,
+    field_id INTEGER NOT NULL,
+    value DATETIME,
+    UNIQUE KEY (bug_id, field_id)
+);
+
+DROP TABLE IF EXISTS cf_selection;
+
+CREATE TABLE cf_selection (
+    bug_id INTEGER,
+    field_id INTEGER NOT NULL,
+    label_id INTEGER,
+    INDEX (bug_id, field_id)
+);
+
+DROP TABLE IF EXISTS cf_transactions;
+
+CREATE TABLE cf_transactions (
+    transaction_id INTEGER AUTO_INCREMENT PRIMARY KEY,
+    bug_id INTEGER NOT NULL,
+    user_id MEDIUMINT NOT NULL,
+    time DATETIME NOT NULL,
+    INDEX (bug_id)
+);
+
+DROP TABLE IF EXISTS cf_integer_activity;
+
+CREATE TABLE cf_integer_activity (
+    transaction_id INTEGER NOT NULL,
+    field_id INTEGER NOT NULL,
+    old_value INTEGER,
+    new_value INTEGER,
+    INDEX (transaction_id)
+);
+
+DROP TABLE IF EXISTS cf_shortstring_activity;
+
+CREATE TABLE cf_shortstring_activity (
+    transaction_id INTEGER NOT NULL,
+    field_id INTEGER NOT NULL,
+    old_value TINYTEXT,
+    new_value TINYTEXT,
+    INDEX (transaction_id)
+);
+
+DROP TABLE IF EXISTS cf_longstring_activity;
+
+CREATE TABLE cf_longstring_activity (
+    transaction_id INTEGER NOT NULL,
+    field_id INTEGER NOT NULL,
+    old_value text,
+    new_value text,
+    INDEX (transaction_id)
+);
+
+DROP TABLE IF EXISTS cf_date_activity;
+
+CREATE TABLE cf_date_activity (
+    transaction_id INTEGER NOT NULL,
+    field_id INTEGER NOT NULL,
+    old_value date,
+    new_value date,
+    INDEX (transaction_id)
+);
+
+DROP TABLE IF EXISTS cf_selection_activity;
+
+CREATE TABLE cf_selection_activity (
+    transaction_id INTEGER NOT NULL,
+    field_id INTEGER NOT NULL,
+    added integer,
+    removed integer,
+    INDEX (transaction_id)
+);
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/editcustomfields.cgi bugzilla-2.17.4-new/editcustomfields.cgi
--- bugzilla-2.17.4/editcustomfields.cgi	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/editcustomfields.cgi	Tue Jul  8 15:49:11 2003
@@ -0,0 +1,1668 @@
+#!/usr/bonsaitools/bin/perl -w
+
+use lib '.';
+
+use Bugzilla::CustomFields qw(
+	read_default_field_data read_selection_metadata classify_fields
+	insert_field_data read_cf_data class_for_type table_for_type form_to_sql
+);
+use Bugzilla::CustomFields::Util;
+use strict;
+use vars qw(%FORM %MFORM $vars $template);
+
+require 'CGI.pl';
+
+#  The keys of %delta are the valid values for the various *-dir CGI
+#  parameters.  The corresponding values are the appropriate offsets for
+#  purposes of splicing.  For example, if @fields is an ordered array
+#  of fields in some group, $base is the offset of the field relative to
+#  which a new field $newfield is to be inserted, and $dir is 'before' or
+#  'after', then the new field would be inserted into the proper place in
+#  the field list by a statement like the following:
+#
+#  splice @fields, $base + $delta{$dir}, 0, $newfield;
+
+my %delta = (
+    before => 0,
+    after  => 1,
+);
+
+#  The keys of %type are the valid values for the various *-type CGI
+#  parameters.
+
+my %type = map { $_ => undef } qw(string selection integer date);
+
+#  %contextsub is a dispatch table.  The keys are the valid values for
+#  the "context" CGI parameter.
+
+my %contextsub = (
+    'main-menu'    => \&handle_main_menu,
+    'edit-field'   => \&edit_field,
+    'rename-field' => \&rename_field,
+    'create-field' => \&show_create_field,
+    'remove-group' => \&remove_group,
+    'remove-field' => \&remove_field,
+);
+
+#  %actionsub is a dispatch table.  The keys are the various values of
+#  the "action" CGI parameter on the main menu page.
+
+my %actionsub = (
+    'Edit'   => \&show_edit,
+    'Create' => \&show_create_field,
+    'Insert' => \&insert_group,
+    'Move'   => \&move_item,
+    'Rename' => \&show_rename,
+    'Remove' => \&show_remove,
+);
+
+############
+#
+#  Main program begins.
+#
+
+ConnectToDatabase();
+
+quietly_check_login();
+
+eval {
+    main();
+};
+
+if ($@) {
+    print "Content-Type: text/html\n\n";
+    ThrowTemplateError($@);
+}
+
+exit;
+
+#
+#  Main program ends.
+#
+############
+
+#
+#  main()
+#
+#  Top-level dispatch subroutine.  If the "context" parameter is present,
+#  dispatches to the appropriate subroutine based on it.  Otherwise
+#  the main menu is shown if one of the "product", "group", or "global"
+#  parameters is present, or the choose-scope page otherwise.
+#
+
+sub main {
+    my $context = $FORM{'context'};
+
+    if (!$context) {
+        if (exists $FORM{'product'}
+                || exists $FORM{'group'}
+                || exists $FORM{'global'}) {
+            show_main_menu();
+        } else {
+            show_choose_scope();
+        }
+        return;
+    }
+
+    defined(my $sub = $contextsub{$context})
+        or die qq(Invalid context "$context"\n);
+
+    $sub->();
+
+}
+
+#
+#  process_template($name)
+#
+#  Prints a minimal CGI header, then processes and prints the template
+#  in template/en/default/custom-field called "$name.html.tmpl".  Throws
+#  a template error if the processing fails.
+#
+
+sub process_template {
+    my $name = shift;
+
+    print "Content-Type: text/html\n\n";
+
+    $template->process("custom-field/$name.html.tmpl", $vars)
+        or ThrowTemplateError($template->error());
+}
+
+#
+#  read_form(@param)
+#
+#  Many routines need to know whether they're being invoked in the context
+#  of a product or a group, and what the numeric value of the "product"
+#  or "group" parameter is.  This routine returns a list containing first,
+#  the string 'product' or 'group'; next, the value of the "product" or
+#  "group" form parameter; next, the form parameters named by @param.
+#
+
+sub read_form {
+    my @param = @_;
+    my $scope;
+
+    if (exists $FORM{'product'}) {
+        $scope = 'product';
+    } elsif (exists $FORM{'group'}) {
+        $scope = 'group';
+    } elsif (exists $FORM{'global'}) {
+        die qq(Global scope not yet implemented\n);
+    } else {
+        die qq(No scope specified\n);
+    }
+
+    return ($scope, @FORM{$scope, @param});
+
+}
+
+#
+#  handle_main_menu()
+#
+#  This routine simply dispatches to the appropriate action subroutine
+#  depending on which submit button was pressed on the main menu.
+#
+
+sub handle_main_menu {
+    my $action = $FORM{'action'};
+
+    defined(my $sub = $actionsub{$action})
+        or die qq(Invalid action "$action"\n);
+
+    $sub->();
+
+}
+
+#
+#  show_choose_scope()
+#
+#  This routine sets up the context for, and then displays, the
+#  choose-scope template.
+#
+
+sub show_choose_scope {
+    my (@products, @groups);
+
+    process_query(qq{
+        SELECT id, name
+        FROM products
+        ORDER BY name
+    }, sub {
+        push @products, { id => $_[0], name => $_[1] };
+    });
+
+    process_query(qq{
+        SELECT group_id, group_name
+        FROM cf_groups
+        WHERE group_name IS NOT NULL
+        ORDER BY group_name
+    }, sub {
+        push @groups, { id => $_[0], name => $_[1] };
+    });
+
+    $vars->{products} = \@products;
+    $vars->{groups}   = \@groups;
+
+    process_template('choose-scope');
+
+}
+
+#
+#  show_main_menu()
+#
+#  This routine calls read_product() or read_group() based on whether
+#  the "product" or "group" CGI parameters is set, then displays the
+#  main-menu template.
+#
+
+sub show_main_menu {
+    if (exists $FORM{'product'}) {
+        read_product($FORM{'product'});
+    } elsif (exists $FORM{'group'}) {
+        read_group($FORM{'group'});
+    } elsif (exists $FORM{'global'}) {
+        die qq(Global scope not yet implemented\n);
+    } else {
+        die qq(Unknown scope\n);
+    }
+    process_template('main-menu');
+}
+
+#
+#  insert_group()
+#
+#  This routine inserts a named group into a product.  The product is
+#  identified by the 'product' CGI parameter, and the group by the
+#  'insert-target' parameter; it will be inserted either before or after
+#  the group identified by the 'insert-base' parameter, based on the
+#  'insert-dir' parameter.  The main menu is redisplayed if all went well.
+#
+
+sub insert_group {
+    my ($scope, $scope_id, $target, $dir, $base) =
+        read_form('insert-target', 'insert-dir', 'insert-base');
+
+    $scope eq 'product' or die qq(Groups may only be inserted into products\n);
+
+    is_numeric($scope_id) or die qq(Invalid product ID "$scope_id"\n);
+    is_numeric($target) or die qq(Invalid group ID "$target"\n);
+    is_numeric($base) or die qq(Invalid group ID "$base"\n);
+
+    defined(my $delta = $delta{$dir})
+        or die qq(Invalid parameter "insert-dir"\n);
+
+    # lock_tables(...);
+
+    product_exists($scope_id) or die qq(Unknown product "$scope_id"\n);
+    group_exists($target)     or die qq(Unknown group "$target"\n);
+
+    my $order = reorder_members($scope => $scope_id);
+
+    defined(my $base_index = $order->{index}->($base))
+        or die qq(Group $base is not a member of product $scope_id\n);
+    !defined($order->{index}->($target))
+        or die qq(Product $scope_id already contains group $target\n);
+
+    $order->{insert}->($base_index + $delta, $target);
+
+    send_sql(qq{
+        INSERT INTO cf_product_members (product_id, group_id, sortkey)
+                                VALUES ($scope_id, $target, 0)
+    });
+
+    $order->{commit}->();
+
+    my @bugs = simple_query(qq{
+        SELECT b.bug_id
+        FROM bugs b
+        WHERE b.product_id = $scope_id
+    });
+
+    if (@bugs) {
+        my @fields = read_group_data($target);
+        my %by_class = classify_fields(@fields);
+        while (my ($class, $fields) = each %by_class) {
+			my @fldid  = map $$_{id},      @$fields;
+			my @values = map $$_{default}, @$fields;
+            insert_field_data($class, \@fldid, \@values, @bugs);
+        }
+    }
+
+    # unlock_tables();
+
+    show_main_menu();
+
+}
+
+#
+#  move_item()
+#
+#  This routine moves the field identified by the CGI parameter
+#  'move-target' to the 'move-dir' side of the field 'move-base'.  It then
+#  redisplays the main menu.
+#
+
+sub move_item {
+    my ($scope, $id, $target, $dir, $base) =
+        read_form('move-target', 'move-dir', 'move-base');
+
+    is_numeric($id) or die qq(Invalid $scope "$id"\n);
+
+    defined(my $delta = $delta{$dir})
+        or die qq(Invalid parameter "move-dir"\n);
+
+    my $order = reorder_members($scope => $id);
+
+    defined(my $source = $order->{index}->($target))
+        or die "No such field $target\n";
+    defined(my $dest = $order->{index}->($base))
+        or die "No such field $base\n";
+
+    ++$source if $source >= $dest;
+
+    $order->{insert}->($dest + $delta, $target);
+    $order->{remove}->($source);
+    $order->{commit}->();
+
+    show_main_menu();
+
+}
+
+#
+#  show_edit()
+#
+#  This routine sets up the context for, and then displays, the edit-field
+#  template.
+#
+
+sub show_edit {
+    my ($scope, $scope_id, $target) = read_form('edit-target');
+
+    is_numeric($scope_id) or die qq(Invalid $scope "$scope_id"\n);
+    is_numeric($target)   or die qq(Invalid field ID "$target"\n);
+
+    if ($scope eq 'product') {
+        product_exists($scope_id)
+            or die qq(Unknown product "$scope_id"\n);
+    } elsif ($scope eq 'group') {
+        group_exists($scope_id)
+            or die qq(Unknown group "$scope_id"\n);
+    }
+
+    field_exists($target) or die qq(Unknown field "$target"\n);
+
+    my $field = Bugzilla::CustomFields::read_custom_fields($target);
+
+    if ($field->{is_selection}) {
+        die qq(Editing of selection fields not yet implemented\n);
+    }
+
+    $vars->{scope}    = $scope;
+    $vars->{scope_id} = $scope_id;
+    $vars->{field}    = $field;
+
+    process_template('edit-field');
+
+}
+
+#
+#  edit_field()
+#
+#  This routine is called when the administrator clicks the "Change Field"
+#  button on the edit-field page.  It resets the default value of the field
+#  identified by the "field" CGI parameter to the value of the
+#  "new-default" parameter.  An exception is thrown if the new default
+#  value isn't properly formatted.  Otherwise, the main menu is
+#  redisplayed.
+#
+
+sub edit_field {
+    my ($field, $default) = @FORM{'field', 'new-default'};
+
+    is_numeric($field) or die qq(Invalid field ID "$field"\n);
+
+    my $type = simple_query(qq{
+        SELECT field_type
+        FROM custom_fields
+        WHERE field_id = $field
+    });
+    defined $type or die qq(Unknown field ID "$field"\n);
+
+	my $table = table_for_type($type);
+
+	$default = form_to_sql($default, $type);
+
+    send_sql(qq{
+        UPDATE $table
+        SET value = $default
+        WHERE bug_id IS NULL AND field_id = $field
+    }); 
+
+    show_main_menu();
+
+}
+
+#
+#  show_rename()
+#
+#  This routine sets up the context for, and then displays, the
+#  rename-field template.
+#
+
+sub show_rename {
+    my ($scope, $scope_id, $field_id) = read_form('rename-target');
+    my $field_name;
+
+    is_numeric($field_id) or die qq(Invalid field ID "$field_id"\n);
+    is_numeric($scope_id) or die qq(Invalid $scope ID "$scope_id"\n);
+
+    process_query(qq{
+        SELECT display_name
+        FROM custom_fields
+        WHERE field_id = $field_id
+    }, sub {
+        $field_name = $_[0];
+    }) or die qq(No such field "$field_id"\n);
+
+    $vars->{field_name} = $field_name;
+    $vars->{field_id}   = $field_id;
+    $vars->{scope}      = $scope;
+    $vars->{scope_id}   = $scope_id;
+
+    process_template('rename-field');
+
+}
+
+#
+#  rename_field()
+#
+#  This routine sets the name of the field identified by the "target-field"
+#  CGI parameter to the value of the "new-name" parameter, then redisplays
+#  the main menu.
+#
+
+sub rename_field {
+    my ($field, $name) = @FORM{'target-field', 'new-name'};
+
+    is_numeric($field)   or die qq(Invalid field ID "$field"\n);
+    field_exists($field) or die qq(Unknown field ID "$field"\n);
+
+    # lock_tables(...);
+
+    valid_display_name($name)
+        or die qq(Invalid field display name "$name"\n);
+
+    $name = sql_quote($name);
+
+    send_sql(qq{
+        UPDATE custom_fields
+        SET display_name = $name
+        WHERE field_id = $field
+    });
+
+    # unlock_tables();
+
+    show_main_menu();
+
+}
+
+#
+#  valid_field_name($name)
+#
+#  Returns true if $name is a valid field name, and false otherwise.  A
+#  valid field name starts with a non-numeric word character, and is
+#  followed by zero or more word characters.
+#
+
+sub valid_field_name {
+    my $name = shift;
+    $name =~ /^[^\W\d]\w*$/;
+}
+
+#
+#  valid_display_name($name)
+#
+#  Returns true if $name is a valid display name, and false otherwise.  A
+#  valid display name does not begin or end with whitespace, and does not
+#  contain any vertical whitespace.
+#
+
+sub valid_display_name {
+    my $name = shift;
+    $name !~ /^\s/ && $name !~ /\s$/ && $name !~ /[\n\r]/;
+}
+
+#
+#  validate_initial_create_params($param)
+#
+#  This routine validates the 'create-type', 'create-dir', 'create-base',
+#  and 'progress' CGI parameters, which are present throughout the field
+#  creation process.  The parameters are stored in the hash referenced by
+#  the $param argument, under the keys 'type', 'dir', 'base', and
+#  'progress', respectively.  In addition, $param->{scope} is set to either
+#  'product' or 'group', and $param->{scope_id} is set to the value of the
+#  'product' or 'group' CGI parameter.
+#
+
+sub validate_initial_create_params {
+    my $param = shift;
+    my ($scope, $scope_id, $type, $dir, $base, $progress) =
+        read_form('create-type', 'create-dir', 'create-base', 'progress');
+
+    exists $type{$type}
+        or die qq(Invalid parameter "create-type"\n);
+    !defined $dir || exists $delta{$dir}
+        or die qq(Invalid parameter "create-dir"\n);
+    !defined $base || is_numeric($base)
+        or die qq(Invalid parameter "create-base"\n);
+
+    if ($progress eq "") {
+        $progress = 0;
+    } elsif ($progress !~ /^\d$/) {
+        die qq(Invalid parameter "progress"\n);
+    }
+
+    if ($scope eq 'product') {
+        is_numeric    ($scope_id) or die qq(Invalid product ID "$scope_id"\n);
+        product_exists($scope_id) or die qq(No such product "$scope_id"\n);
+        !defined $base || product_has_group($scope_id, $base)
+            or die qq(No such group $base in product $scope_id\n);
+    } elsif ($scope eq 'group') {
+        is_numeric($scope_id) or die qq(Invalid group ID "$scope_id"\n);
+        group_exists($scope_id) or die qq(No such group "$scope_id"\n);
+        !defined $base || group_has_field($scope_id, $base)
+            or die qq(No such field $base in group $scope_id\n);
+    }
+
+    @$param{qw(scope scope_id progress type dir base)} =
+        ($scope, $scope_id, $progress, $type, $dir, $base);
+
+}
+
+#
+#  validate_secondary_create_params($param)
+#
+#  This routine validates the CGI parameters which appear on the first
+#  create-field page.  These always include 'field-name', 'display-name',
+#  and 'required', and others that depend on the type of the field being
+#  created:  'default' (for integer and date fields), 'with-time' (for
+#  date fields), 'size' (for string fields), and/or 'multi', 'unset', and
+#  'domain' (for selection fields).  If valid, these parameters are stored
+#  in the hash referenced by the $param argument using shorter, more
+#  convenient names.
+#
+
+sub validate_secondary_create_params {
+    my $param = shift;
+    my ($name, $dname, $required) =
+        @FORM{'field-name', 'display-name', 'required'};
+
+    valid_field_name($name)    or die qq(Invalid field name "$name"\n);
+    valid_display_name($dname) or die qq(Invalid display name "$dname"\n);
+    $required = $required ? 1 : 0;
+
+    @$param{qw(name dname required)} = ($name, $dname, $required);
+
+    my $quoted = sql_quote($name);
+    0 == process_query(qq{
+        SELECT 1
+        FROM custom_fields
+        WHERE field_name = $quoted
+    }) or die qq(Custom field "$name" already exists\n);
+
+    if ($param->{type} eq 'integer') {
+        $param->{default} = parse_integer($FORM{'default'});
+    } elsif ($param->{type} eq 'date') {
+        my ($default, $withtime) = @FORM{'default', 'with-time'};
+        $withtime = $withtime ? 1 : 0;
+        $default = stringify_date($default, $withtime);
+        @$param{qw(with_time default)} = ($withtime, $default);
+    } elsif ($param->{type} eq 'string') {
+        my $size = $FORM{'size'};
+        $size eq 'short' || $size eq 'long'
+            or die qq(Invalid parameter "size"\n);
+        $param->{size} = $size;
+    } elsif ($param->{type} eq 'selection') {
+        my ($multi, $unset, $domain) = @FORM{'multi', 'unset', 'domain'};
+        $multi eq 'yes' || $multi eq 'no'
+            or die qq(Invalid parameter "multi"\n);
+        $unset =~ /^[^\r\n]+$/
+            or die qq(Invalid parameter "unset"\n);
+        if ($param->{progress} == 1) {
+            my %seen = ($unset => 1);
+            $domain =~ s/^\s+//gm;
+            $domain =~ s/\s+$//gm;
+            $domain =~ s/\n+/\n/g;
+            $domain !~ /\r/ or die qq(Invalid parameter "domain"\n);
+            $domain =~ /\S/ or die qq(Invalid parameter "domain"\n);
+            $domain = [ grep !$seen{$_}++, split /\n/, $domain ];
+        } else {
+            $domain = [ deserialize_array($domain) ];
+        }
+        @$param{qw(multi unset domain)} = ($multi, $unset, $domain);
+    }
+
+}
+
+#
+#  validate_tertiary_create_params($param)
+#
+#  This routine validates the CGI parameters which appear on the second
+#  create-field page (which is only reached when a string or selection
+#  field is being created).  The only parameter entered on this page in
+#  both cases is 'default', which is stored in $param->{default}.
+#
+
+sub validate_tertiary_create_params {
+    my $param = shift;
+    my $default;
+
+    if ($param->{type} eq 'string') {
+        $default = $FORM{'default'};
+        if ($param->{size} eq 'short') {
+            $default !~ /[\r\n]/
+                or die qq(Invalid short string value "$default"\n);
+        }
+    } elsif ($param->{type} eq 'selection') {
+        my %domain;
+        @domain{ @{ $param->{domain} } } = ();
+        $default = $MFORM{'default'};
+        if ($param->{multi} eq 'no') {
+            @$default == 1 or die qq(Single-selection fields can have ) .
+                                  qq(only a single default value\n);
+            my $label = $default->[0];
+            if ($label eq $param->{unset}) {
+                $default = undef;
+            } else {
+                exists $domain{$label}
+                    or die qq(Invalid default selection value "$label"\n);
+            }
+        } else {
+            for my $value (@$default) {
+                exists $domain{$value}
+                    or die qq(Invalid default selection value "$value"\n);
+            }
+        }
+    } else {
+        die qq(Invalid parameter "progress"\n);
+    }
+
+    $param->{default} = $default;
+
+}
+
+#
+#  show_create_field()
+#
+#  This routine sets up the context for, and then displays, the
+#  create-field template.  If all of the necessary information to create a
+#  field has been supplied, then the field is created and the main menu is
+#  redisplayed instead.
+#
+
+sub show_create_field {
+    my (%param, %hidden);
+
+    validate_initial_create_params(\%param);
+    validate_secondary_create_params(\%param) if $param{progress} >= 1;
+    validate_tertiary_create_params(\%param)  if $param{progress} >= 2;
+
+    if ($param{progress} == 0) {
+        @hidden{$param{scope}, 'create-type'} = @param{qw(scope_id type)};
+        @hidden{'create-dir', 'create-base'} = @param{qw(dir base)}
+            if defined $param{dir} && defined $param{base};
+        @$vars{qw(hidden create_type progress)} = (\%hidden, $param{type}, 1);
+        process_template('create-field');
+        return;
+    }
+
+    if ($param{progress} == 1) {
+        if ($param{type} eq 'integer' || $param{type} eq 'date') {
+            create_field(\%param);
+            show_main_menu();
+            return;
+        } elsif ($param{type} eq 'string') {
+            @hidden{$param{scope}, qw(
+                create-type field-name display-name required size
+            )} = @param{qw(
+                scope_id type name dname required size
+            )};
+            @hidden{'create-dir', 'create-base'} = @param{qw(dir base)}
+                if defined $param{dir} && defined $param{base};
+            @$vars{qw(hidden create_type size progress)} =
+                (\%hidden, $param{type}, $param{size}, 2);
+            process_template('create-field');
+            return;
+        } elsif ($param{type} eq 'selection') {
+            @hidden{$param{scope}, qw(
+                create-type field-name display-name required multi
+                unset domain
+            )} = (@param{qw(
+                scope_id type name dname required multi unset
+            )}, serialize_array($param{domain}));
+            @hidden{'create-dir', 'create-base'} = @param{qw(dir base)}
+                if defined $param{dir} && defined $param{base};
+            @$vars{qw(hidden create_type multi domain progress)} =
+                (\%hidden, @param{qw(type multi domain)}, 2);
+            $vars->{unset} = $param{unset} if $param{multi} eq 'no';
+            process_template('create-field');
+            return;
+        }
+    }
+
+    if ($param{progress} == 2) {
+        create_field(\%param);
+        show_main_menu();
+        return;
+    }
+    
+    die qq(Invalid parameter "progress"\n);
+
+}
+
+#
+#  product_exists($product)
+#
+#  Returns true if the product identified by the numeric ID $product
+#  exists, and false otherwise.
+#
+
+sub product_exists {
+    my $product = shift;
+    return process_query(qq{
+        SELECT 1
+        FROM products
+        WHERE id = $product
+    });
+}
+
+#
+#  group_exists($group)
+#
+#  Returns true if the group identified by the numeric ID $group exists,
+#  and false otherwise.
+#
+
+sub group_exists {
+    my $group = shift;
+    return process_query(qq{
+        SELECT 1
+        FROM cf_groups
+        WHERE group_id = $group
+    });
+}
+
+#
+#  field_exists($field)
+#
+#  Returns true if the field identified by the numeric ID $field exists,
+#  and false otherwise.
+#
+
+sub field_exists {
+    my $field = shift;
+    return process_query(qq{
+        SELECT 1
+        FROM custom_fields
+        WHERE field_id = $field
+    });
+}
+
+#
+#  product_has_group($product, $group)
+#
+#  Returns true if the product identified by the numeric ID $product
+#  contains a group with numeric ID $group, and false otherwise.
+#
+
+sub product_has_group {
+    my ($product, $group) = @_;
+    return process_query(qq{
+        SELECT 1
+        FROM cf_product_members
+        WHERE product_id = $product AND group_id = $group
+    });
+}
+
+#
+#  group_has_field($group, $field)
+#
+#  Returns true if the group identified by the numeric ID $group contains
+#  the field with numeric ID $field, and false otherwise.
+#
+
+sub group_has_field {
+    my ($group, $field) = @_;
+    return process_query(qq{
+        SELECT 1
+        FROM cf_group_members
+        WHERE group_id = $group AND field_id = $field
+    });
+}
+
+#
+#  serialize_array($aref)
+#
+#  Returns a string containing a serialized representation of the array
+#  referenced by $aref.  "|" characters within each element of the array
+#  are doubled, and each element is then joined together with "|"
+#  characters.  This is useful for storing array quantities (for example,
+#  the "domain" parameter of a selection field) in a hidden CGI element.
+#
+#  Example: serialize_array(['abc', 'd|e', 'fgh']) returns 'abc|d||e|fgh'.
+#
+
+sub serialize_array {
+    my $aref = shift;
+    my @array = @$aref;
+    s/\|/||/g for @array;
+    return join '|', @array;
+}
+
+#
+#  deserialize_array($ser)
+#
+#  Undoes the serialization performed by serialize_array().
+#
+
+sub deserialize_array {
+    my $ser = shift;
+    my $elem = "";
+    my @array;
+    while ($ser =~ /(.*?)\|/gcs) {
+        $elem .= $1;
+        if ($ser =~ /\G\|/gc) {
+            $elem .= '|';
+        } else {
+            push @array, $elem;
+            $elem = "";
+        }
+    }
+    push @array, $elem . substr($ser, pos($ser));
+    return @array;
+}
+
+#
+#  create_field($param)
+#
+#  This routine creates a custom field.  All of the necessary information
+#  must be present in the hash referenced by the $param argument under the
+#  following keys:
+#
+#    'name':      The name of the field.
+#    'dname':     The display name of the field.
+#    'base':      The ID of the field relative to which the new field will be
+#                 placed.
+#    'dir':       Which side of the field 'base' the new field should be
+#                 placed on, either 'before' or 'after'.
+#    'required':  True if the new field is required on field creation,
+#                 false otherwise.
+#    'type':      The major type of the new field, either 'integer',
+#                 'date', 'string', or 'selection'.
+#    'default':   The default value of the field.
+#    'with_time': (date fields only)  True if the field should include a
+#                 time-of-day, false otherwise.
+#    'size':      (string fields only)  'long' if the field should be a
+#                 long string field, 'short' otherwise.
+#    'multi':     (selection fields only)  'yes' if the field should be a
+#                 multiselection field, 'no' otherwise.
+#    'unset':     (selection fields only)  The label that should be used to
+#                 identify a field with no label selected.
+#    'domain':    (selection fields only)  A reference to an array
+#                 containing all of the valid labels for the new field.
+#
+#  All of these values are presumed to have already been validated.
+#
+
+sub create_field {
+    my $param = shift;
+    my ($bugtable, $label_id, $dbtype, $field_id, $newindex, @bugs);
+    my $selection_id = 'NULL';
+
+    my ($scope, $scope_id, $base, $dir, $required, $name, $dname,
+        $type, $default) = @$param{qw(
+        scope scope_id base dir required name dname type default
+    )};
+
+    if ($type eq 'integer') {
+        $dbtype = 'i';
+        $bugtable = 'cf_integer';
+    } elsif ($type eq 'date') {
+        $dbtype = $param->{with_time} ? 't' : 'd';
+        $bugtable = 'cf_date';
+    } elsif ($type eq 'string') {
+        $dbtype = $param->{size} eq 'short' ? 's' : 'l';
+        $bugtable = $dbtype eq 's' ? 'cf_shortstring' : 'cf_longstring';
+    } elsif ($type eq 'selection') {
+        $dbtype = $param->{multi} eq 'no' ? 'e' : 'm';
+        $bugtable = 'cf_selection';
+    }
+
+    # lock_tables( ... );
+
+    my $order = reorder_members($scope => $scope_id);
+
+    if (defined $base && defined $dir) {
+        defined($order->{index}->($base)) or die qq(No such field "$base"\n);
+        $newindex = $order->{index}->($base) + $delta{$dir};
+    } else {
+        $newindex = 0;
+    }
+
+    if ($type eq 'selection') {
+        ($selection_id, $label_id) =
+            create_selection_metadata(@$param{'unset', 'domain'});
+        @$default = @$label_id{ @$default };
+    }
+
+	my $class = class_for_type($dbtype);
+
+    $name   = sql_quote($name);
+    $dname  = sql_quote($dname);
+    $dbtype = sql_quote($dbtype);
+
+    send_sql(qq{
+        INSERT INTO custom_fields (
+            field_id, field_name, display_name, field_type,
+            required, selection_id
+        ) VALUES (
+            NULL, $name, $dname, $dbtype, $required, $selection_id
+        )
+    });
+    $field_id = last_insert_id();
+
+    $order->{insert}->($newindex, $field_id);
+
+    if ($scope eq 'product') {
+        my $group_id = create_singleton_group($field_id);
+        send_sql(qq{
+            INSERT INTO cf_product_members (product_id, group_id, sortkey)
+                                    VALUES ($scope_id, $group_id, 0)
+        });
+        @bugs = simple_query(qq{
+            SELECT bug_id
+            FROM bugs
+            WHERE product_id = $scope_id
+        });
+    } elsif ($scope eq 'group') {
+        send_sql(qq{
+            INSERT INTO cf_group_members (group_id, field_id, sortkey)
+                                  VALUES ($scope_id, $field_id, 0)
+        });
+        @bugs = simple_query(qq{
+            SELECT b.bug_id
+            FROM bugs b
+              INNER JOIN cf_product_members pm USING (product_id)
+            WHERE pm.group_id = $scope_id
+        });
+    }
+
+    $order->{commit}->();
+
+    insert_field_data($class, [ $field_id ], [ $default ], 'NULL', @bugs);
+
+    # unlock_tables();
+
+}
+
+#
+#  create_singleton_group($field_id)
+#
+#  Creates an unnamed group containing the single field identified by the
+#  numeric ID $field_id.  Returns the new group's numeric ID.
+#
+
+sub create_singleton_group {
+    my $field_id = shift;
+
+    send_sql(qq{
+        INSERT INTO cf_groups (group_id, group_name)
+                       VALUES (NULL, NULL)
+    });
+    my $group_id = last_insert_id();
+
+    send_sql(qq{
+        INSERT INTO cf_group_members (group_id, field_id, sortkey)
+                              VALUES ($group_id, $field_id, 1)
+    });
+
+    return $group_id;
+
+}
+
+#
+#  create_selection_metadata($unset, $domain)
+#
+#  Populates the cf_selections and cf_selection_labels tables with
+#  information for a new selection field.  Returns a two-element list.  The
+#  first element is the numeric ID of the new selection set; the second is
+#  a reference to hash which maps the new labels to the numeric ID of each.
+#
+
+sub create_selection_metadata {
+    my ($unset, $domain) = @_;
+    my ($n, $selection_id, %label_id);
+
+    $unset = sql_quote($unset);
+
+    send_sql(qq{
+        INSERT INTO cf_selections (selection_id, unset_label)
+                           VALUES (NULL, $unset)
+    });
+    $selection_id = last_insert_id();
+
+    $n = 1;
+    for my $label (@$domain) {
+        my $quoted = sql_quote($label);
+        send_sql(qq{
+            INSERT INTO cf_selection_labels (
+                label_id, selection_id, sortkey, label
+            ) VALUES (
+                NULL, $selection_id, $n, $quoted
+            )
+        });
+        $label_id{$label} = last_insert_id();
+        ++$n;
+    }
+
+    return ($selection_id, \%label_id);
+
+}
+
+#
+#  show_remove()
+#
+#  Dispatches to either show_remove_from_product or show_remove_from_group,
+#  depending on whether the 'product' or 'group' CGI parameter is present.
+#
+
+sub show_remove {
+    my ($scope, $scope_id, $target) = read_form('remove-target');
+
+    if ($scope eq 'product') {
+        show_remove_from_product($scope_id, $target);
+    } elsif ($scope eq 'group') {
+        show_remove_from_group($scope_id, $target);
+    }
+
+}
+
+#
+#  show_remove_from_product($product, $target)
+#
+#  This routine sets up the context for, and then displays, the
+#  remove-from-product template.  $target is the numeric ID of a group
+#  to be removed from the product identified by the numeric ID $product.
+#
+
+sub show_remove_from_product {
+    my ($product, $target) = @_;
+    my ($product_name, $name, $object);
+
+    is_numeric($product) or die qq(Invalid product "$product"\n);
+    is_numeric($target)  or die qq(Invalid group "$target"\n);
+
+    process_query(qq{
+        SELECT name
+        FROM products
+        WHERE id = $product
+    }, sub {
+        $product_name = $_[0];
+    }) or die qq(No such product "$product"\n);
+
+    process_query(qq{
+        SELECT group_name
+        FROM cf_groups
+        WHERE group_id = $target
+    }, sub {
+        $name = $_[0];
+    }) or die qq(Unknown group "$target"\n);
+
+    if (defined $name) {
+        $object = 'group';
+    } else {
+        $object = 'field';
+        process_query(qq{
+            SELECT cf.display_name
+            FROM custom_fields cf
+              INNER JOIN cf_group_members gm USING (field_id)
+            WHERE gm.group_id = $target
+        }, sub {
+            $name = $_[0];
+        }) or die qq(Unknown field "$target"\n); #  DB corrupt!
+    }
+
+    $vars->{bugs_affected} = simple_query(qq{
+        SELECT COUNT(*)
+        FROM bugs
+        WHERE product_id = $product
+    });
+
+    $vars->{product}      = $product;
+    $vars->{product_name} = $product_name;
+    $vars->{target_name}  = $name;
+    $vars->{target_id}    = $target;
+    $vars->{object}       = $object;
+
+    process_template('remove-from-product');
+
+}
+
+#
+#  show_remove_from_group($group_id, $field_id)
+#
+#  This routine sets of the context for, and then displays, the
+#  remove-from-group template.  $field_id is the numeric ID of a field to
+#  be removed from the group identified by the numeric ID $group_id.
+#
+
+sub show_remove_from_group {
+    my ($group_id, $field_id) = @_;
+    my ($group_name, $field_name, $bug_count, $product_count);
+
+    is_numeric($group_id) or die qq(Invalid group ID "$group_id"\n);
+    is_numeric($field_id) or die qq(Invalid field ID "$field_id"\n);
+
+    process_query(qq{
+        SELECT group_name
+        FROM cf_groups
+        WHERE group_id = $group_id
+    }, sub {
+        $group_name = $_[0];
+    }) or die qq(No such group ID "$group_id"\n);
+
+    defined $group_name or die qq(Cannot remove group "$group_id"\n);
+    #  (Because it's private to some product.)
+
+    process_query(qq{
+        SELECT display_name
+        FROM custom_fields
+        WHERE field_id = $field_id
+    }, sub {
+        $field_name = $_[0];
+    }) or die qq(No such field ID "$field_id"\n);
+
+    $bug_count = simple_query(qq{
+        SELECT COUNT(*)
+        FROM bugs
+          INNER JOIN cf_product_members pm USING (product_id)
+        WHERE pm.group_id = $group_id
+    });
+
+    $product_count = simple_query(qq{
+        SELECT COUNT(*)
+        FROM cf_product_members
+        WHERE group_id = $group_id
+    });
+
+    $vars->{group}         = $group_id;
+    $vars->{group_name}    = $group_name;
+    $vars->{target_id}     = $field_id;
+    $vars->{target_name}   = $field_name;
+    $vars->{bugs_affected} = $bug_count;
+    $vars->{product_count} = $product_count;
+
+    process_template('remove-from-group');
+
+}
+
+sub lock_tables {
+    # ...
+    return;
+}
+
+sub unlock_tables {
+    send_sql("UNLOCK TABLES");
+}
+
+#
+#  remove_bug_data($bugs, $meta)
+#
+#  This routines deletes all custom field data associated with each of the
+#  bugs whose numeric IDs are contained in the array referenced by the
+#  $bugs argument.  $meta is a field metadata object, as returned by
+#  read_field_metadata().
+#
+
+sub remove_bug_data {
+    my ($bugs, $meta) = @_;
+    return if @$bugs == 0;
+    my $bug_set = join(',', @$bugs);
+
+    my @trans = simple_query(qq{
+        SELECT transaction_id
+        FROM cf_transactions
+        WHERE bug_id IN ($bug_set)
+    });
+    my $trans_set = join(',', @trans);
+
+    while (my ($table, $fields) = $meta->{each_table}->()) {
+        my $field_set = join(',', @$fields);
+        send_sql(qq{
+            DELETE FROM $table
+            WHERE bug_id IN ($bug_set)
+              AND field_id IN ($field_set)
+        });
+        if (@trans) {
+            send_sql(qq{
+                DELETE FROM ${table}_activity
+                WHERE transaction_id IN ($trans_set)
+            });
+        }
+    }
+
+    if (@trans) {
+        send_sql(qq{
+            DELETE FROM cf_transactions
+            WHERE transaction_id IN ($trans_set)
+        });
+    }
+
+}
+
+#
+#  remove_field_metadata($group, $meta)
+#
+#  This routine removes all metadata describing each of the fields
+#  described by $meta (a field metadata object as returned by
+#  read_field_metadata()) from the group identified by the numeric ID
+#  $group.  The remaining fields in that group are re-ordered to close the
+#  gap.
+#
+
+sub remove_field_metadata {
+    my ($group, $meta) = @_;
+    my @fields;
+
+    while (my ($table, $fields) = $meta->{each_table}->()) {
+        push @fields, @$fields;
+        my $field_set = join(',', @$fields);
+        send_sql(qq{
+            DELETE FROM $table
+            WHERE bug_id IS NULL
+              AND field_id IN ($field_set)
+        });
+        if ($table eq 'cf_selection') {
+            my $selection_set = join(',', $meta->{selection_id}->(@$fields));
+            send_sql(qq{
+                DELETE FROM cf_selection_labels
+                WHERE selection_id IN ($selection_set)
+            });
+            send_sql(qq{
+                DELETE FROM cf_selections
+                WHERE selection_id IN ($selection_set)
+            });
+        }
+    }
+
+    my $order = reorder_members(group => $group);
+
+    #  Sort into sortkey-order:
+    @fields = sort { $order->{index}->($a) <=> $order->{index}->($b) } @fields;
+
+    for (my $i = 0; $i < @fields; ++$i) {
+        $order->{remove}->($order->{index}->($fields[$i]) - $i);
+    }
+
+    $order->{commit}->();
+
+    my $field_set = join(',', @fields);
+
+    send_sql(qq{
+        DELETE FROM cf_group_members
+        WHERE group_id = $group
+          AND field_id IN ($field_set)
+    });
+
+    send_sql(qq{
+        DELETE FROM custom_fields
+        WHERE field_id IN ($field_set)
+    });
+
+}
+
+#
+#  remove_field()
+#
+#  This routine removes all bug data and field metadata for the field
+#  identified by the numeric ID $FORM{'target'} from the group identified
+#  by the numeric ID $FORM{'group'}.  It then redisplays the main menu.
+#
+
+sub remove_field {
+    my ($group, $field) = @FORM{'group', 'target'};
+
+    is_numeric($group) or die qq(Invalid group ID "$group"\n);
+    is_numeric($field) or die qq(Invalid field ID "$field"\n);
+
+    #  lock_tables(...);
+
+    my @bugs = simple_query(qq{
+        SELECT b.bug_id
+        FROM bugs b
+          INNER JOIN cf_product_members pm USING (product_id)
+          INNER JOIN cf_group_members gm USING (group_id)
+        WHERE gm.field_id = $field
+    });
+
+    my $meta = read_field_metadata($field);
+
+    remove_bug_data(\@bugs, $meta);
+
+    remove_field_metadata($group, $meta);
+
+    # unlock_tables();
+
+    show_main_menu();
+
+}
+
+#
+#  remove_group()
+#
+#  This routine removes the group identified by the numeric ID
+#  $FORM{'target'} from the product identified by the numeric ID
+#  $FORM{'product'}.  The remaining groups in that product are re-ordered
+#  to close the gap.  Bug data for all of the fields in the group is
+#  removed.  If the group is anonymous--that is, if it contains only a
+#  single field which belongs to a single product--the contained field is
+#  removed as well.  Finally, the main menu is redisplayed.
+#
+
+sub remove_group {
+    my ($product, $group) = @FORM{'product', 'target'};
+    my $group_name;
+
+    is_numeric($product) or die qq(Invalid product ID "$product"\n);
+    is_numeric($group)   or die qq(Invalid group ID "$group"\n);
+
+    product_exists($product) or die qq(Unknown product ID "$product"\n);
+    group_exists($group)     or die qq(Unknonw group ID "$group"\n);
+
+    # lock_tables(...);
+
+    is_numeric($group) or die qq(Invalid group ID "$group"\n);
+
+    process_query(qq{
+        SELECT group_name
+        FROM cf_groups
+        WHERE group_id = $group
+    }, sub {
+        $group_name = $_[0];
+    }) or die qq(Unknown group "$group"\n);
+
+    my $order = reorder_members(product => $product);
+    $order->{remove}->($order->{index}->($group));
+    $order->{commit}->();
+
+    my @bugs = simple_query(qq{
+        SELECT b.bug_id
+        FROM bugs b
+          INNER JOIN cf_product_members pm USING (product_id)
+        WHERE pm.group_id = $group
+    });
+
+    my @fields = simple_query(qq{
+        SELECT field_id
+        FROM cf_group_members
+        WHERE group_id = $group
+    });
+
+    my $meta = read_field_metadata(@fields);
+
+    remove_bug_data(\@bugs, $meta);
+
+    send_sql(qq{
+        DELETE FROM cf_product_members
+        WHERE product_id = $product
+          AND group_id = $group
+    });
+
+    if (!defined $group_name) {
+        remove_field_metadata($group, $meta);
+        send_sql(qq{
+            DELETE FROM cf_groups
+            WHERE group_id = $group
+        });
+    }
+
+    # unlock_tables();
+
+    show_main_menu();
+
+}
+
+#
+#  read_group($group_id)
+#
+#  Sets up the context for the main menu template in the case that a group
+#  (identified by the numeric ID $group_id) is to be displayed.
+#
+
+sub read_group {
+    my $group_id = shift;
+
+    my $group_name = simple_query(qq{
+        SELECT group_name
+        FROM cf_groups
+        WHERE group_id = $group_id
+    });
+
+    my @fields = read_group_data($group_id);
+
+    read_selection_metadata(@fields);
+
+    $vars->{scope}           = 'group';
+    $vars->{scope_id}        = $group_id;
+    $vars->{scope_name}      = $group_name;
+    $vars->{custom_fields}   = \@fields;
+    $vars->{editable_fields} = \@fields;
+
+}
+
+#
+#  read_group_data($group_id)
+#
+#  Creates a field hash for every field in the group identified by the
+#  numeric ID $group_id.  Returns a list containing all of the newly-created
+#  field hashes.
+#
+
+sub read_group_data {
+    my $group_id = shift;
+    my @fields;
+
+    send_sql(qq{
+        SELECT cf.field_id, cf.field_name, cf.display_name, cf.field_type,
+               cf.required, cf.selection_id
+        FROM custom_fields cf
+          INNER JOIN cf_group_members gm USING (field_id)
+        WHERE gm.group_id = $group_id
+        ORDER BY gm.sortkey
+    });
+
+    while (MoreSQLData()) {
+        push @fields, read_cf_data();
+    }
+
+    read_default_field_data(@fields);
+
+    return @fields;
+
+}
+
+#
+#  read_product($product_id)
+#
+#  Sets up the context for the main menu template in the case that a product
+#  (idenfitied by the numeric ID $product_id) is to be displayed.
+#
+
+sub read_product {
+    my $product_id = shift;
+    my ($last_name, $current_group, @fields, @groups);
+
+    my $product_name = simple_query(qq{
+        SELECT name
+        FROM products
+        WHERE id = $product_id
+    });
+
+    my $product_clause = defined $product_id
+        ? "product_id = $product_id"
+        : "product_id IS NULL";
+
+    #  The left joins here are necessary because we still want to know
+    #  when a product contains a group even if the group itself has no
+    #  members:
+
+    send_sql(qq{
+        SELECT cf.field_id, cf.field_name, cf.display_name, cf.field_type,
+               cf.required, cf.selection_id, g.group_id, g.group_name
+        FROM cf_groups g
+          INNER JOIN cf_product_members pm USING (group_id)
+          LEFT JOIN cf_group_members gm USING (group_id)
+          LEFT JOIN custom_fields cf USING (field_id)
+        WHERE pm.$product_clause
+        ORDER BY pm.sortkey, gm.sortkey
+    });
+
+    while (MoreSQLData()) {
+        my ($field, $group_id, $group_name) = read_cf_data();
+
+        push @fields, $field if $field;
+
+        unless (defined $last_name
+                && defined $group_name
+                && $last_name eq $group_name) {
+            $current_group = [ $field ? $field : () ];
+            push @groups, {
+                id => $group_id,
+                name => $group_name,
+                members => $current_group
+            };
+        } else {
+            push @$current_group, $field;
+        }
+
+        $last_name = $group_name;
+
+    }
+
+    read_default_field_data(@fields);
+    read_selection_metadata(@fields);
+
+    my @editable = map @{ $$_{members} }, grep !defined $$_{name}, @groups;
+
+    my @other_groups;
+    my @group_ids = map $$_{id}, grep defined $$_{name}, @groups;
+    my $group_id_clause;
+    if (@group_ids) {
+        my $group_set = join(',', @group_ids);
+        $group_id_clause = "AND group_id NOT IN ($group_set)";
+    } else {
+        $group_id_clause = "";
+    }
+
+    process_query(qq{
+        SELECT group_id, group_name
+        FROM cf_groups
+        WHERE group_name IS NOT NULL $group_id_clause
+        ORDER BY group_name
+    }, sub {
+        push @other_groups, { id => $_[0], name => $_[1] };
+    });
+
+    $vars->{scope}               = 'product';
+    $vars->{scope_id}            = $product_id;
+    $vars->{scope_name}          = $product_name;
+    $vars->{custom_fields}       = \@fields;
+    $vars->{editable_fields}     = \@editable;
+    $vars->{custom_field_groups} = \@groups;
+    $vars->{other_groups}        = \@other_groups;
+
+}
+
+#
+#  reorder_members($scope, $id)
+#
+#  This routine returns an object which can be used to modify the members of
+#  a product (if $scope is 'product') or group (if $scope is 'group').  $id
+#  is the numeric ID of the product or group which is to be modified.
+#
+#  The returned object is a hash of subroutines, so one calls method "foo"
+#  on it like this:  $object->{foo}->(@args).  The valid methods are:
+#
+#  index($id)
+#  Returns the numeric index (starting at zero) of the group or field with
+#  numeric ID $id.
+#
+#  remove($n)
+#  Removes the group or field located at index $n.
+#
+#  insert($n, $id)
+#  Inserts a new group or field with numeric ID $id before the existing
+#  element at index $n.
+#
+#  commit()
+#  Updates the database to reflect the new group or field order.  No further
+#  methods should be called on the object once commit() has been called.
+#
+
+sub reorder_members {
+    my ($scope, $id) = @_;
+    my ($table, $column, $key, @member, %index, %sortkey);
+
+    if ($scope eq 'product') {
+        $table  = 'cf_product_members';
+        $column = 'group_id';
+        $key    = 'product_id';
+    } elsif ($scope eq 'group') {
+        $table  = 'cf_group_members';
+        $column = 'field_id';
+        $key    = 'group_id';
+    } else {
+        die qq(Internal error in reorder_members: unknown scope "$scope"\n);
+    }
+
+    process_query(qq{
+        SELECT $column, sortkey
+        FROM $table
+        WHERE $key = $id
+        ORDER BY sortkey
+    }, sub {
+        push @member, $_[0];
+        $sortkey{ $_[0] } = $_[1];
+    });
+
+    @index{ @member } = 0 .. $#member;
+
+    return {
+        index  => sub { return $index{ $_[0] } },
+        remove => sub { splice @member, $_[0], 1 },
+        insert => sub { splice @member, $_[0], 0, $_[1] },
+        commit => sub {
+            my @when;
+            for (my $i = 0; $i < @member; ++$i) {
+                if ($sortkey{ $member[$i] } != $i + 1) {
+                    push @when, "WHEN $member[$i] THEN " . ($i + 1);
+                }
+            }
+            if (@when) {
+                my $case_body = join(' ', @when) . ' ELSE sortkey';
+                send_sql(qq{
+                    UPDATE $table
+                    SET sortkey = CASE $column $case_body END
+                    WHERE $key = $id
+                });
+            }
+        }
+    };
+
+}
+
+#
+#  read_field_metadata(@fields)
+#
+#  Returns an object that can be used to obtain information about each field
+#  whose numeric ID appears in @fields.  The object is a
+#  hash-of-subroutines, like the one returned by reorder_members().
+#
+#  each_table()
+#  Iterates over each table which stores data for @fields.  Called in scalar
+#  context, returns the name of one of these tables; in list context,
+#  returns a second scalar, a reference to an array containing the field
+#  IDs of each field for which data is stored in the table named by the
+#  first element.
+#
+#  selection_id(@field_id)
+#  Returns a list containing the selection IDs for each field in @field_ids.
+#  @field_ids should be a subset of the fields initially passed to
+#  read_field_metadata().
+#
+
+sub read_field_metadata {
+    my @fields = @_;
+    my (%by_table, %selection_id);
+
+    if (@fields) {
+        my $field_set = join(',', @fields);
+        process_query(qq{
+            SELECT field_id, field_type, selection_id
+            FROM custom_fields
+            WHERE field_id IN ($field_set)
+        }, sub {
+            my ($field_id, $type, $sel_id) = @_;
+			my $table = table_for_type($type);
+            push @{ $by_table{$table} }, $field_id;
+            $selection_id{$field_id} = $sel_id;
+        });
+    }
+
+    return {
+        each_table   => sub { each %by_table },
+        selection_id => sub { @selection_id{ @_ } }
+    };
+
+}
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/enter_bug.cgi bugzilla-2.17.4-new/enter_bug.cgi
--- bugzilla-2.17.4/enter_bug.cgi	Tue Jul  8 12:45:20 2003
+++ bugzilla-2.17.4-new/enter_bug.cgi	Tue Jul  8 12:45:24 2003
@@ -37,6 +37,7 @@
 use lib qw(.);
 
 use Bugzilla::Constants;
+use Bugzilla::CustomFields qw(get_custom_fields);
 require "CGI.pl";
 
 use vars qw(
@@ -356,6 +357,8 @@
 $vars->{'group'} = \@groups;
 
 $vars->{'default'} = \%default;
+
+$vars->{'custom_fields'} = [ get_custom_fields($product_id) ];
 
 $vars->{'use_keywords'} = 1 if (@::legal_keywords);
 
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/post_bug.cgi bugzilla-2.17.4-new/post_bug.cgi
--- bugzilla-2.17.4/post_bug.cgi	Tue Jul  8 12:45:20 2003
+++ bugzilla-2.17.4-new/post_bug.cgi	Tue Jul  8 12:45:24 2003
@@ -27,6 +27,9 @@
 use lib qw(.);
 
 use Bugzilla::Constants;
+use Bugzilla::CustomFields qw(
+    get_custom_fields check_custom_fields store_custom_fields
+);
 require "CGI.pl";
 
 use Bug;
@@ -182,6 +185,9 @@
 CheckFormFieldDefined(\%::FORM, 'bug_file_loc');
 CheckFormFieldDefined(\%::FORM, 'comment');
 
+my @fields = get_custom_fields($product_id);
+check_custom_fields(@fields);
+
 my @used_fields;
 foreach my $field (@bug_fields) {
     if (exists $::FORM{$field}) {
@@ -407,6 +413,8 @@
 # Get the bug ID back.
 SendSQL("select LAST_INSERT_ID()");
 my $id = FetchOneColumn();
+
+store_custom_fields($id, @fields);
 
 # Add the group restrictions
 foreach my $grouptoadd (@groupstoadd) {
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/process_bug.cgi bugzilla-2.17.4-new/process_bug.cgi
--- bugzilla-2.17.4/process_bug.cgi	Tue Jul  8 12:45:20 2003
+++ bugzilla-2.17.4-new/process_bug.cgi	Tue Jul  8 12:45:24 2003
@@ -42,6 +42,10 @@
 # Use the Flag module to modify flag data if the user set flags.
 use Bugzilla::Flag;
 
+use Bugzilla::CustomFields qw(
+    get_custom_fields check_custom_fields update_custom_fields
+);
+
 # Shut up misguided -w warnings about "used only once":
 
 use vars qw(%versions
@@ -73,9 +77,13 @@
 # For each bug being modified, make sure its ID is a valid bug number 
 # representing an existing bug that the user is authorized to access.
 my @idlist;
+my @custom_fields;
+
 if (defined $::FORM{'id'}) {
     ValidateBugID($::FORM{'id'});
     push @idlist, $::FORM{'id'};
+    my $product_id = get_product_id($::FORM{'product'});
+    @custom_fields = get_custom_fields($product_id, $::FORM{'id'});
 } else {
     foreach my $i (keys %::FORM) {
         if ($i =~ /^id_([1-9][0-9]*)/) {
@@ -519,6 +527,7 @@
                    \@{$::versions{$::FORM{'product'}}});
     CheckFormField(\%::FORM, 'op_sys', \@::legal_opsys);
     CheckFormFieldDefined(\%::FORM, 'longdesclength');
+    check_custom_fields(@custom_fields);
 }
 
 my $action = '';
@@ -1098,7 +1107,9 @@
             "keyworddefs READ, groups READ, attachments READ, " .
             "group_control_map AS oldcontrolmap READ, " .
             "group_control_map AS newcontrolmap READ, " .
-            "group_control_map READ");
+            "group_control_map READ, " .
+            "cf_integer $write, cf_date $write, cf_shortstring $write, " .
+            "cf_longstring $write, cf_selection $write");
     my @oldvalues = SnapShotBug($id);
     my %oldhash;
     # Fun hack.  @::log_columns only contains the component_id,
@@ -1310,6 +1321,7 @@
     
     if ($::comma ne "") {
         SendSQL($query);
+        update_custom_fields($::FORM{'id'}, @custom_fields) if @custom_fields;
     }
 
     # Check for duplicates if the bug is [re]open
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/bug/create/create.html.tmpl bugzilla-2.17.4-new/template/en/default/bug/create/create.html.tmpl
--- bugzilla-2.17.4/template/en/default/bug/create/create.html.tmpl	Thu Apr 24 13:49:42 2003
+++ bugzilla-2.17.4-new/template/en/default/bug/create/create.html.tmpl	Tue Jul  8 11:19:54 2003
@@ -193,6 +193,8 @@
     </td>
   </tr>
 
+  [% INCLUDE "custom-field/show-fields.html.tmpl" %]
+
   [% IF UserInGroup('editbugs') %]
     [% IF use_keywords %]
       <tr>
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/bug/edit.html.tmpl bugzilla-2.17.4-new/template/en/default/bug/edit.html.tmpl
--- bugzilla-2.17.4/template/en/default/bug/edit.html.tmpl	Wed Apr  9 02:29:19 2003
+++ bugzilla-2.17.4-new/template/en/default/bug/edit.html.tmpl	Tue Jul  8 11:20:33 2003
@@ -247,6 +247,10 @@
     </td>
   </tr>
 
+  [% INCLUDE "custom-field/show-fields.html.tmpl"
+             show_value = 1
+             custom_fields = bug.custom_fields %]
+
   [% IF Param('usestatuswhiteboard') %] 
     <tr>
       <td align="right">
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/choose-scope.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/choose-scope.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/choose-scope.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/choose-scope.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,53 @@
+[% PROCESS global/header.html.tmpl
+   title = "Choose a scope"
+%]
+
+[% IF products.size > 0 %]
+  <form action="editcustomfields.cgi" method="get">
+  <table>
+    <tr>
+      <td>
+        <input type="submit" value="Choose a product:">
+      </td>
+      <td>
+        <select name="product">
+          [% FOREACH product = products %]
+            <option value="[% product.id FILTER html %]">
+              [% product.name FILTER html %]
+            </option>
+          [% END %]
+        </select>
+      </td>
+    </tr>
+  </table>
+  </form>
+[% END %]
+
+[% IF groups.size > 0 %]
+  <form action="editcustomfields.cgi" method="get">
+  <table>
+    <tr>
+      <td>
+        <input type="submit" value="Choose a group:">
+      </td>
+      <td>
+        <select name="group">
+          [% FOREACH group = groups %]
+            <option value="[% group.id FILTER html %]">
+              [% group.name FILTER html %]
+            </option>
+          [% END %]
+        </select>
+      </td>
+    </tr>
+  </table>
+  </form>
+[% END %]
+
+<form action="editcustomfields.cgi" method="get">
+<input type="hidden" name="global" value="1">
+<input type="submit" value="Examine the global scope">
+</form>
+
+[% PROCESS global/footer.html.tmpl %]
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/create-field.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/create-field.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/create-field.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/create-field.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,122 @@
+[% PROCESS global/header.html.tmpl
+   title = "New Field"
+%]
+
+<form action="editcustomfields.cgi" method="post">
+
+<input type="hidden" name="context" value="create-field">
+<input type="hidden" name="progress" value="[% progress %]">
+[% FOREACH param = hidden.keys.sort %]
+  <input type="hidden" name="[% param %]" value="[% hidden.$param | html %]">
+[% END %]
+
+[% IF progress == 1 %]
+  <p>
+    New field name:
+    <input type="text" name="field-name">
+  </p>
+  <p>
+    New field display name:
+    <input type="text" name="display-name">
+  </p>
+  <p>
+    <input type="checkbox" name="required">
+    Field is required on bug creation
+  </p>
+[% END %]
+
+[% SWITCH create_type %]
+[% CASE 'integer' %]
+  <p>
+    Default value:
+    <input type="text" name="default">
+  </p>
+  <p>
+    <input type="submit" value="Create integer field">
+  </p>
+[% CASE 'date' %]
+  <p>
+    <input type="checkbox" name="with-time">
+    Include time-of-day
+  </p>
+  <p>
+    Default value:
+    <input type="text" name="default">
+  </p>
+  <p>
+    <input type="submit" value="Create date field">
+  </p>
+[% CASE 'string' %]
+  [% IF progress == 1 %]
+    <p>
+      <input type="radio" name="size" value="short">
+      Short string (single line, up to 255 characters)
+      <br>
+      <input type="radio" name="size" value="long">
+      Long string (multiline, up to 64K characters)
+    </p>
+    <p>
+      <input type="submit" value="Continue field creation">
+    </p>
+  [% ELSIF progress == 2 %]
+    <p>
+      Default value:
+      [% IF size == 'short' %]
+        <input type="text" name="default">
+      [% ELSE %]
+        <textarea name="default" rows="5" cols="40"></textarea>
+      [% END %]
+    </p>
+    <p>
+      <input type="submit" value="Create string field">
+    </p>
+  [% END %]
+[% CASE 'selection' %]
+  [% IF progress == 1 %]
+    <p>
+      <input type="radio" name="multi" value="no">
+      Single selection
+      <br>
+      <input type="radio" name="multi" value="yes">
+      Multiple selection
+    </p>
+    <p>
+      Unset label:
+      <input type="text" name="unset" value="None">
+    </p>
+    <p>
+      Value set:
+      <textarea name="domain" rows="5" cols="40"></textarea>
+    </p>
+    <p>
+      <input type="submit" value="Continue field creation">
+    </p>
+  [% ELSIF progress == 2 %]
+    <p>
+      Default value:
+      [% IF multi == 'no' %]
+        <select name="default">
+          [%# What if unset label equals some other label???  %]
+          <option value="[% unset | html %]">[% unset | html %]</option>
+          [% FOREACH label = domain %]
+            <option value="[% label | html %]">[% label | html %]</option>
+          [% END %]
+        </select>
+      [% ELSE %]
+        <select name="default" multiple size="7">
+          [% FOREACH label = domain %]
+            <option value="[% label | html %]">[% label | html %]</option>
+          [% END %]
+        </select>
+      [% END %]
+    </p>
+    <p>
+      <input type="submit" value="Create selection field">
+    </p>
+  [% END %]
+[% END %]
+
+</form>
+
+[% PROCESS global/footer.html.tmpl %]
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/edit-field.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/edit-field.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/edit-field.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/edit-field.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,31 @@
+[% PROCESS global/header.html.tmpl
+   title = "Main Menu"
+%]
+
+<form action="editcustomfields.cgi" method="post">
+
+<input type="hidden" name="context" value="edit-field">
+<input type="hidden" name="[% scope %]" value="[% scope_id %]">
+<input type="hidden" name="field" value="[% field.id %]">
+
+<p>
+  Enter the field's new default value:
+
+[% IF field.is_integer
+      or field.is_date
+      or (field.is_string and not field.extended) %]
+  <input type="text" name="new-default" value="[% field.default | html %]">
+[% ELSIF field.is_string %]
+  <textarea name="new-default" rows="5" cols="40">[% field.default | html %]</textarea>
+[% END %]
+
+</p>
+
+<p>
+  <input type="submit" value="Change field">
+</p>
+
+</form>
+
+[% PROCESS global/footer.html.tmpl %]
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/main-menu.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/main-menu.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/main-menu.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/main-menu.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,262 @@
+[% PROCESS global/header.html.tmpl
+   title = "Main Menu"
+   style = "
+     table.custom-field-table th, table.custom-field-table td {
+       vertical-align: top;
+       padding: 5;
+     }
+   "
+%]
+
+<p>
+  This is
+  [% IF scope == 'product' %]
+    product
+  [% ELSIF scope == 'group' %]
+    group
+  [% END %]
+  <b>[% scope_name FILTER html %]</b>.
+</p>
+
+[% IF scope == 'product';
+      numrows = custom_field_groups.size;
+      IF editable_fields.size == numrows;
+         has_group_column = 0;
+      ELSE;
+         has_group_column = 1;
+      END;
+   ELSE;
+      numrows = custom_fields.size;
+      has_group_column = 0;
+   END
+%]
+
+<table border class="custom-field-table">
+  <tr>
+    [% IF has_group_column %]<th></th>[% END %]
+    <th>Name</th>
+    <th>Display Name</th>
+    <th>Type</th>
+    <th>Required?</th>
+    <th>Default</th>
+  </tr>
+  [% IF scope == 'product' %]
+    [% FOREACH group = custom_field_groups %]
+      [% IF group.members.size == 0 %]
+        <tr>
+          <td>[% PROCESS group_link %]</td>
+          <td colspan="5" align="center">
+            <i>There are no fields in this group.</i>
+          </td>
+        </tr>
+      [% ELSE %]
+        [% FOREACH field = group.members %]
+          <tr>
+            [% IF has_group_column and loop.index == 0 %]
+              <td rowspan="[% group.members.size %]">
+                [% IF group.name.defined %]
+                  [% PROCESS group_link %]
+                [% END %]
+              </td>
+            [% END %]
+            [% PROCESS field_row %]
+          </tr>
+        [% END %]
+      [% END %]
+    [% END %]
+  [% ELSE %]
+    [% FOREACH field = custom_fields %]
+      <tr>
+        [% PROCESS field_row %]
+      </tr>
+    [% END %]
+  [% END %]
+</table>
+
+<hr>
+
+<form method="post" action="editcustomfields.cgi">
+
+<input type="hidden" name="context" value="main-menu">
+<input type="hidden" name="[% scope %]" value="[% scope_id %]">
+
+[% IF editable_fields.size > 0 %]
+  <p>
+    <input type="submit" name="action" value="Edit">
+    the custom field&nbsp;
+    [% PROCESS cf_select_fields name="edit-target" %].
+  </p>
+[% END %]
+
+<p>
+  <input type="submit" name="action" value="Create">
+  a new&nbsp;
+  [% PROCESS cf_types name="create-type" %]
+  custom field[% IF numrows == 0 %].[% ELSE %]&nbsp;
+  [% PROCESS cf_direction name="create-dir" %]&nbsp;
+  [% PROCESS cf_select_groups name="create-base" %].[% END %]
+</p>
+
+[% IF scope == 'product' and other_groups.size > 0 %]
+  <p>
+    <input type="submit" name="action" value="Insert">
+    the field group&nbsp;
+    [% PROCESS cf_select_other_groups name="insert-target" %][% IF numrows == 0 %].[% ELSE %]&nbsp;
+    [% PROCESS cf_direction name="insert-dir" dir="before" %]&nbsp;
+    [% PROCESS cf_select_groups name="insert-base" %].[% END %]
+  </p>
+[% END %]
+
+[% IF numrows > 1 %]
+  <p>
+    <input type="submit" name="action" value="Move">&nbsp;
+    [% PROCESS cf_select_groups name="move-target" %]
+    so that it is&nbsp;
+    [% PROCESS cf_direction name="move-dir" dir="after" %]&nbsp;
+    [% INCLUDE cf_select_groups name="move-base" select=2 %].
+  </p>
+[% END %]
+
+[% IF editable_fields.size > 0 %]
+  <p>
+    <input type="submit" name="action" value="Rename">
+    the custom field&nbsp;
+    [% PROCESS cf_select_fields name="rename-target" %].
+  </p>
+[% END %]
+
+[% IF numrows > 0 %]
+  <p>
+    <input type="submit" name="action" value="Remove">
+    the field or group&nbsp;
+    [% PROCESS cf_select_groups name="remove-target" %].
+  </p>
+[% END %]
+
+</form>
+
+[% PROCESS global/footer.html.tmpl %]
+
+
+
+
+[% BLOCK field_row %]
+  <td>[% PROCESS custom_name         %]</td>
+  <td>[% PROCESS custom_display_name %]</td>
+  <td>[% PROCESS custom_type         %]</td>
+  <td align="center">[% PROCESS custom_required %]</td>
+  <td align="center">[% PROCESS custom_default  %]</td>
+[% END %]
+
+[% BLOCK custom_name %]
+  [% field.name FILTER html %]
+[% END %]
+
+[% BLOCK custom_display_name %]
+  [% field.display_name FILTER html %]
+[% END %]
+
+[% BLOCK custom_type %]
+  [% IF field.is_integer %]
+    integer
+  [% ELSIF field.is_date %]
+    date [% IF field.extended %] (with time) [% END %]
+  [% ELSIF field.is_string %]
+    string [% IF field.extended %] (long) [% ELSE %] (short) [% END %]
+  [% ELSIF field.is_selection %]
+    selection [% IF field.extended %] (multiple) [% ELSE %] (single) [% END %]
+  [% END %]
+[% END %]
+
+[% BLOCK custom_required %]
+  [% IF field.required %] Yes [% ELSE %] No [% END %]
+[% END %]
+
+[% BLOCK custom_default %]
+  [% IF not field.is_selection %]
+    [% IF not field.default.defined %]
+      <i>None<i>
+    [% ELSE %]
+      <pre>[% '"' IF field.is_string %][% field.default FILTER html %][% '"' IF field.is_string %]</pre>
+    [% END %]
+  [% ELSE %]
+    <select size="3" [%- " multiple" IF field.extended %]>
+      [% IF not field.extended %]
+        <option [%- " selected" IF not field.default.defined %]>
+          [% field.unset FILTER html %]
+        </option>
+      [% END %]
+      [% FOREACH value = field.domain %]
+        <option [%- " selected" IF field.default.contains(value) %]>
+          [% value FILTER html %]
+        </option>
+      [% END %]
+    </select>
+  [% END %]
+[% END %]
+
+[% BLOCK cf_select_fields %]
+  <select name="[% name FILTER html %]">
+    [% FOREACH field = editable_fields %]
+      <option value="[% field.id %]">
+        [% field.display_name FILTER html %]
+      </option>
+    [% END %]
+  </select>
+[% END %]
+
+[% BLOCK cf_select_groups %]
+  <select name="[% name FILTER html %]">
+    [% IF scope == 'product' %]
+      [% FOREACH group = custom_field_groups %]
+        <option [%- " selected" IF loop.index + 1 == select %] value="[% group.id %]">
+          [% IF group.name.defined %]
+            [% group.name FILTER html %]
+          [% ELSE %]
+            [% group.members.0.display_name FILTER html %]
+          [% END %]
+        </option>
+      [% END %]
+    [% ELSE %]
+      [% FOREACH field = editable_fields %]
+        <option value="[% field.id %]">
+          [% field.display_name FILTER html %]
+        </option>
+      [% END %]
+    [% END %]
+  </select>
+[% END %]
+
+[% BLOCK cf_select_other_groups %]
+  <select name="[% name FILTER html %]">
+    [% FOREACH group = other_groups %]
+      <option value="[% group.id %]">[% group.name FILTER html %]</option>
+    [% END %]
+  </select>
+[% END %]
+
+[% BLOCK cf_direction %]
+  <select name="[% name FILTER html %]">
+    <option [%- " selected" IF dir == "before" %] value="before">
+      before
+    </option>
+    <option [%- " selected" IF dir == "after" %] value="after">
+      after
+    </option>
+  </select>
+[% END %]
+
+[% BLOCK cf_types %]
+  <select name="[% name FILTER html %]">
+    <option value="string">string</option>
+    <option value="selection">selection</option>
+    <option value="date">date</option>
+    <option value="integer">integer</option>
+  </select>
+[% END %]
+
+[% BLOCK group_link %]
+  <a href="editcustomfields.cgi?group=[% group.id %]">
+    [% group.name FILTER html %]
+  </a>
+[% END %]
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/remove-from-group.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/remove-from-group.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/remove-from-group.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/remove-from-group.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,35 @@
+[% PROCESS global/header.html.tmpl
+   title = "Remove a Field"
+%]
+
+<form action="editcustomfields.cgi" method="post">
+
+<input type="hidden" name="context" value="remove-field">
+<input type="hidden" name="group" value="[% group %]">
+<input type="hidden" name="target" value="[% target_id %]">
+
+<p>
+  You are about to remove the field
+  <b>[% target_name FILTER html %]</b> from the group
+  <b>[% group_name FILTER html %]</b>.  This will permanently remove
+  data from [% bugs_affected %]
+  [% IF bugs_affected == 1 %] bug [% ELSE %] bugs [% END %]
+  in
+  [% product_count %]
+  [% IF product_count == 1 %] product. [% ELSE %] products. [% END %]
+</p>
+
+<p>
+  Press the following button to confirm that you really want to perform
+  this operation.
+</p>
+
+<input type="submit" value="Remove Field">
+
+</form>
+
+
+
+
+[% PROCESS global/footer.html.tmpl %]
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/remove-from-product.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/remove-from-product.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/remove-from-product.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/remove-from-product.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,37 @@
+[% PROCESS global/header.html.tmpl
+   title = "Remove a Group"
+%]
+
+<form action="editcustomfields.cgi" method="post">
+
+<input type="hidden" name="context" value="remove-group">
+<input type="hidden" name="product" value="[% product %]">
+<input type="hidden" name="target" value="[% target_id %]">
+
+<p>
+  You are about to remove the [% object %]
+  <b>[% target_name FILTER html %]</b> from the product
+  <b>[% product_name FILTER html %]</b>.  This will
+  [% IF bugs_affected == 0 %]
+    not affect any bugs.
+  [% ELSIF bugs_affected == 1 %]
+    permanently remove data from one bug.
+  [% ELSE %]
+    permanently remove data from [% bugs_affected %] bugs.
+  [% END %]
+</p>
+
+<p>
+  Press the following button to confirm that you really want to perform
+  this operation.
+</p>
+
+<input type="submit" value="Remove [% object %]">
+
+</form>
+
+
+
+
+[% PROCESS global/footer.html.tmpl %]
+
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/rename-field.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/rename-field.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/rename-field.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/rename-field.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,25 @@
+[% PROCESS global/header.html.tmpl
+   title = "Rename a Field"
+%]
+
+<form action="editcustomfields.cgi" method="post">
+
+<input type="hidden" name="context" value="rename-field">
+<input type="hidden" name="[% scope %]" value="[% scope_id %]">
+<input type="hidden" name="target-field" value="[% field_id %]">
+
+<p>
+  Rename the field <b>[% field_name FILTER html %]</b> to:
+  <input type="text" name="new-name">.
+</p>
+
+<p>
+  <input type="submit" value="Rename Field">
+</p>
+
+</form>
+
+
+
+
+[% PROCESS global/footer.html.tmpl %]
diff -Naur --exclude=.htaccess --exclude=localconfig --exclude=data bugzilla-2.17.4/template/en/default/custom-field/show-fields.html.tmpl bugzilla-2.17.4-new/template/en/default/custom-field/show-fields.html.tmpl
--- bugzilla-2.17.4/template/en/default/custom-field/show-fields.html.tmpl	Wed Dec 31 16:00:00 1969
+++ bugzilla-2.17.4-new/template/en/default/custom-field/show-fields.html.tmpl	Tue Jul  8 11:17:07 2003
@@ -0,0 +1,33 @@
+[% DEFAULT show_value = 0 # otherwise, show default %]
+
+[% FOREACH field = custom_fields %]
+  [% IF show_value; value = field.value; ELSE; value = field.default; END %]
+  <tr>
+    <td>
+      <b>[% field.display_name | html %]:</b>
+    </td>
+    <td colspan="3">
+      [% IF field.is_selection %]
+        <select [% ' multiple size="5"' IF field.extended %] name="[% field.name | html %]">
+          [% IF not field.extended %]
+            <option [% " selected" IF value.size == 0 %] value="[% field.unset | html %]">[% field.unset | html %]</option>
+          [% END %]
+          [% FOREACH label = field.domain %]
+            <option [% " selected" IF value.contains(label) %] value="[% label | html %]">[% label | html %]</option>
+          [% END %]
+        </select>
+      [% ELSE %]
+        [% value = field.as_string(value) %]
+        [% IF field.is_integer
+              or field.is_date
+              or (field.is_string and not field.extended) %]
+          <input type="text" name="[% field.name | html %]" value="[% value | html %]">
+        [% ELSE %]
+          <textarea name="[% field.name | html %]" rows="5" cols="40">[% value | html %]</textarea>
+        [% END %]
+      [% END %]
+    </td>
+  </tr>
+[% END %]
+
+
